#!/usr/local/bin/perl -w
#
# Given a full text XML file, generate the dependency graphs.
# Usage:
#
# use dg;
#
# ...
# 
# ProcessDG($Filename, \&callback, $debug);
# 
# sub callback {
#     my($topnode, $errorflag) = @_;
#     ...
# }
#
# ProcessDG() processes each sentence in the XML file $Filename and
# calls the user-supplied subroutine callback() with the top node
# of the dependency graph and a flag indicating if the sentence parsed
# correctly. For now, $errorflag is 0 if there were no errors, and 1
# otherwise. Later, we'll want to add fatal vs. non-fatal, and possibly
# error codes.
#
# If $debug is provided and > 0, generate snapshots before
# each <annotationSet>.
#
#				NOTES
#
#     Frame Processing
#
# When you see an annotationSet, you create $FrameNode immediately
# (with no parents). When you see the Target, it is moved under
# the FrameNode, and the FrameNode's parent is set to the Target's
# old parent. When labels in layers are seen, non-local child edges
# are always created to the given span and the labels are inserted.
#
# Once the whole sentence has been parsed, there's a post-processing
# step, which includes moving non-locals references up as far in the tree
# as they can go, moving local nodes to the appropriate position,
# eliminating redundant nodes, and assigning phrase-type and headedness.
#
# See the file SynSemDocumentation.txt (and the comments) for more detail.
#
# Discontinuous targets are now correctly implemented.
#   Example: There still are ...
#            -----       ---
#
#				 TODO
#
# Redivvy components rationally into utils, grammar, mutaphrase, parsing, xml parsing
#
# Hash-based phrase-structure grammar
#
# Handle conjunctions.  (This will wait until implementation of the hash-based psg)
#
# Handle null instantiations.
#
#
#
#				 BUGS
#
# The <text> MUST appear before any <annotationSet>.
#
# Given the large number of circular references, the code probably
# leaks quite a lot.
#
# The code should probably be more OO.
#
# The @AllNodes, @AllEdges, %StartToWord, %EndToWord, and %SymToRef
# are probably not needed.
#
            
use strict;

use XML::Parser;
use Data::Dumper;
use FileHandle;
use IPC::Open2;
use FileHandle;

# Globals used by the XML parser and in building the DG

# Globals for XML parsing

my $Document='';    # The document the example occurs in
my $Paragraph=0;    # The paragraph number within the document
my $SentNum=0;      # The sentence number within the paragraph
my $Text;	  # Current sentence text
my $InText;	  # If 1, currently in a <text>...</text>
my $LayerName;	  # E.g. Target, FE, GF, PT
my $LabelName;	  # E.g. Target, Dep, PP, Document, etc.

my $GovNode;      # The current "Gov" node.
my $XNode;        # The current "X" node, if any.
# my $GovEdge;	  # The non-local edge going (originally) from the gov to the frame.
my %TargetType;   # Hash from symbol of a node to its type: Target, Supp, or Cop.
my %TargetEdgeType; # Hash from symbol of a node to the original parent edgetype, before that was converted to Target, Supp, or Cop.
my $TargetError=0;# Flags insidious errors involving the target that require deleting the entire anno-set 

# Globals for synsem tree entities

my $SentenceNode; # The top node of the DG for the current sentence
my $FrameNode;	  # The node for a frame we're currently working on.
my @AllNodes;     # Simple list of all nodes
my %StartToWord;  # Hash of start position => node ref (for words only)
my %EndToWord;
my %SymToRef;	  # Hash of symbolic name to reference (edge or node).
my $SymIndex;	  # Used to generate symbolic names

# Caller communication

my $Callback;	  # User-supplied subroutine to call with top sentence node
my $Debug;	  # If > 0, generate snapshots before each annotationSet.
my $ErrorFlag;    # Initialized to zero, set to one if problems found, returned to caller.

# Semantic pertinentia

# This hash maps from the names of frames with transparent-noun targets to the FE
# which contains the SemHead for that frame. 

my %TransparentFE=('Aggregate'=>'Individuals', 'Type'=>'Category', 'Quantity'=>'Individuals',
		'Part_edge'=>'Whole', 'Part_inner_outer'=>'Whole', 'Part_ordered_segments'=>'Whole', 
		'Part_orientational'=>'Whole', 'Part_piece'=>'Substance', 'Part_whole'=>'Whole', 'Rest'=>'Whole', 
		'Measure_duration'=>'Process', 'Measure_area'=>'Area', 'Measure_by_action'=>'Entity', 
		'Measure_linear_extent'=>'Stuff', 'Measure_mass'=>'Stuff', 'Measure_volume'=>'Stuff',
		'Cardinal_numbers'=>'Entity');

# Fillers of the following FEs are generally interpreted as relational, and should not be marker phrases.

my %RelEdges;

map { $RelEdges{$_}=1; } qw(Source Path Goal Place Time);

# Unfortunately, it is not the case that all of the FEs with the above names are relational.
# As a hack, I have implemented the following hash which stores the exceptions per FE, per frame.

my %RelExceptions;

map { $RelExceptions{Goal}{$_}=1; } qw(Accomplishment Assistance Attempt Attempting_scenario
					   Military Project Purpose Resolve_attempt_scenario 
					   Successful_action Success_or_failure);




# Grammar pertinentia

my $POS_TAGGER = "/u/chays/tree-tagger/bin/tree-tagger -quiet /u/chays/tree-tagger/lib/english.par";

my %Punct=map {$_=>"$_"} qw#' ` `` '' ( ) : ,#;

my %NonHeadPOStoPTs=(DT=>'NP', POS=>'Poss', CC=>'None', map {$_=>'None'} keys %Punct);

my %HeadPOStoPTs=(NNS=>['NP', 'Poss'], 
		  NPS=>['NP', 'Poss'],
		  NN=>['NP', 'N', 'Poss'], 
		  NP =>['NP', 'N', 'Poss'], 
		  PP=>['NP'],
		  WP=>['NP'],
		  LS=>['Quant'],
		  CD=>['Quant'],
		  DT=>['None'],
		  'PP\$'=>['Poss'], 
		  'WP\$'=>['Poss'],
		  'MD'=>['Sfin', 'VPfin', 'Srel', 'Swhether', 'Sinterrog', '<s>'],
		  VBZ=>['Sfin', 'VPfin', 'Srel', 'Swhether', 'Sinterrog', '<s>'],		  
		  VHZ=>['Sfin', 'VPfin', 'Srel', 'Swhether', 'Sinterrog', '<s>'],		  
		  VVZ=>['Sfin', 'VPfin', 'Srel', 'Swhether', 'Sinterrog', '<s>'],
		  VBD=>['Sfin', 'VPfin', 'Srel', 'Swhether', 'Sinterrog', '<s>'],
		  VHD=>['Sfin', 'VPfin', 'Srel', 'Swhether', 'Sinterrog', '<s>'],
		  VVD=>['Sfin', 'VPfin', 'Srel', 'Swhether', 'Sinterrog', '<s>'],
		  VBP=>['Sfin', 'VPfin', 'Srel', 'Swhether', 'Sinterrog', '<s>'],
		  VHP=>['Sfin', 'VPfin', 'Srel', 'Swhether', 'Sinterrog', '<s>'],
		  VVP=>['Sfin', 'VPfin', 'Srel', 'Swhether', 'Sinterrog', '<s>'],
		  VBI=>['VPbrst', 'Sbrst', 'Sinterrog', 'Swhether', 'Sforto'],
		  VHI=>['VPbrst', 'Sbrst', 'Sinterrog', 'Swhether', 'Sforto'],
		  VVI=>['VPbrst', 'Sbrst', 'Sinterrog', 'Swhether', 'Sforto'],
		  VB=>['VPbrst', 'Sbrst', 'Sinterrog', 'Swhether', 'Sforto'],
		  VH=>['VPbrst', 'Sbrst', 'Sinterrog', 'Swhether', 'Sforto'],
		  VV=>['VPbrst', 'Sbrst', 'Sinterrog', 'Swhether', 'Sforto'],
		  VBG=>['VPing', 'Sing'],
		  VHG=>['VPing', 'Sing'],
		  VVG=>['VPing', 'Sing'],
		  VBN=>['VPed'],
		  VHN=>['VPed'],
		  VVN=>['VPed'],
		  JJ=>['AJP','A'],
		  JJR=>['AJP','A'],
		  JJS=>['AJP','A'],
		  RB=>['AVP'],
		  RBR=>['AVP'],
		  RBS=>['AVP'],
		  RP=>['AVP'],
		  WRB=>['AVP'],
		  'IN'=>['PP', 'Sub', 'PPing', 'PPinterrog'],
		  'TO'=>['PP', 'PPing', 'PPinterrog', 'VPto'],
		  'POS'=>['None'],
		  'EX'=>['NP'],
		  'PUQ'=>['QUO'],
		  "``"=>['QUO'],
		  "`"=>['QUO'],
		  "'"=>['QUO'],
		  "''"=>['QUO']);

my %PTtoHeadPOS=InvertHash(%HeadPOStoPTs);

my %PTtoType= (Sfin=>'V', VPfin=>'V', VPbrst=>'V', VPing=>'V', VPed=>'V',
	       'Srel'=>'V', 'Swhether'=>'V', 'Sinterrog'=>'V', '<s>'=>'V',
	       Sbrst=>'V', Sforto=>'V', VPto=>'V',
	       NP=>'N', N=>'N', Poss=>'N',
	       PP=>'Prep', Sub=>'Prep', PPing=>'Prep', PPinterrog=>'Prep',
	       AJP=>'A', A=>'A', 
	       AVP=>'Adv',
	       Quant=>'Num');

my %MarkerToBare=(Poss=>'NP', PPing=>'VPing', Sub=>'Sfin', PPinterrog=>'Sinterrog', 
		  VPto=>'VPbrst', Sfin=>'Sfin', PP=>'NP');

my %MarkerToHeadPOS=(PP=>['TO', 'IN', 'PRF', 'PRP'], PPinterrog=>['TO', 'IN', 'PRF', 'PRP'], PPing=>['TO', 'IN', 'PRF', 'PRP'], 
		     Sfin=>['IN','CJT'], VPto=>['TO', 'TO0'], Sub=>['IN', 'CJS'], Poss=>['POS']);

# The following ranks word-types in terms of their need to be heads.
# This is especially useful when hierarchicizing Supps/Cops/Govs
# and when trying to less randomly guess which words of an undifferentiated,
# PT-less phrase might be the head.

my %TypeRanking = (Adv=>'0', Num=>'0', A=>'1', N=>'2', Prep=>'3', V=>'4');

# The first of the following two hashes encodes GFs which are special, i.e. have
# an effect on the PT of their parents. The second gives a value for PT/GF pairs: 1
# means GF is required, .5 for GF is allowed.

my %GFtoCheck=(Ext=>1, Gen=>1);

my %PTrequiresGF=(Sfin=>{'Ext'=>1}, 
		  Sing=>{'Ext'=>1}, 
		  Sforto=>{'Ext'=>1}, 
		  Sbrst=>{Ext=>1}, 
		  '<s>'=>{'Ext'=>1}, 
		  Srel=>{Ext=>1}, 
		  NP=>{'Gen'=>.5}, 
		  Sinterrog=>{'Ext'=>.5}, 
		  Swhether=>{Ext=>.5});

my %FormsToGFs = ('I'=>'Ext', 'my'=>'Gen',
		  'your'=>'Gen', 
		  'he'=>'Ext', 'his'=>'Gen', 
		  'she'=>'Ext',
		  'its'=>'Gen',
		  'we'=>'Ext', 'our'=>'Gen',
		  'they'=>'Ext', 'their'=>'Gen',
		  'thou'=>'Ext', 'thy'=>'Gen');

my %CopulaWordForms=(VBFS=>['am', "'m"],
		     VBP=>['are', "'re"],
		     VB=>['be'],
		     VBDS=>['was'],
		     VBDP=>['were'],
		     VBN=>['been'],
		     VBG=>['being'],
		     VBZ=>['is', "'s"] );

my %Aux=('will'=>'will', 'would'=>'would', 
	 'am'=>'be', "'m"=>'be', 'are'=>'be', "'re"=>'be', 'be'=>'be', 'was'=>'be', 'were'=>'be', 'been'=>'be', 'being'=>'be', 'is'=>'be', "'s"=>'be',
	 'have'=>'have', 'has'=>'have', 'had'=>'have', 'having'=>'have', 'hast'=>'have',
	 'do'=>'do', 'does'=>'do', 'did'=>'do', 'done'=>'do', 'doing'=>'do', 'dost'=>'do');

my %Conjunction=('and'=>'and', 'or'=>'or', 'but'=>'but');

######################################################################

sub InvertHash {
    my %inputHash=(@_);
    my %outputHash;
    foreach my $key (keys %inputHash) {
	if (ref($inputHash{$key}) eq 'ARRAY') {
	    foreach my $value (@{$inputHash{$key}}) {
		$outputHash{$value}=[] unless (exists($outputHash{$value}));
		push(@{$outputHash{$value}}, $key) unless (FindInList($key, $outputHash{$value}));
	    }
	} else {
	    $outputHash{$inputHash{$key}}=[] unless (exists($outputHash{$inputHash{$key}}));
	    push(@{$outputHash{$inputHash{$key}}}, $key) unless (FindInList($outputHash{$inputHash{$key}}, $key));
	}
    } 
    return %outputHash;
}


######################################################################
#
# Given a full-text XML file, iteratively create the dependency graph
# for each sentence. See FTEndTag() for where the output is handled.
#

sub ProcessDGs {
    my($filename, $callback, $debug) = @_;
    my($parser);
    $Callback = $callback;
    if (!defined($debug)) {
	$debug = 0;
    }
    $Debug = $debug;
    $parser = new XML::Parser(Handlers => { Start => \&FTStartTag,
					    End   => \&FTEndTag,
					    Char  => \&FTText});
    $parser->parsefile($filename);

    # All the work occurs in FTStartTag, FTEndTag, and FTText.
    # The actual callback happens in FTEndTag().

} # ProcessDGs()

sub FTStartTag {
    my($expat, $elem, %attr) = @_;
    my($node);

    # Handle <text>...</text>
    if ($elem eq 'text') {
	$InText = 1;
	$Text = '';

    } elsif ($elem eq 'document') {
	$Document=$attr{description} if (exists($attr{description}));
	$Paragraph=0;
    } elsif ($elem eq 'paragraph') {
	$Paragraph=$attr{documentOrder} if (exists($attr{documentOrder}));
	$SentNum=0;

    # Initialize all globals at <sentence> (except the purely positional $Document, 
    # $Paragraph, and $SentNum), since each sentence is processed independently.

    } elsif ($elem eq 'sentence') {
	@AllNodes = ();

	%StartToWord = ();
	%EndToWord = ();
	$SymIndex = 0;
	$FrameNode = undef;
	$SentenceNode = MakeSimpleNode('List');
	$SentenceNode->{PT} = '<s>';
	$SentenceNode->{ID} = $attr{ID};
	$SentenceNode->{Context}=join(', ', $Document, "Paragraph $Paragraph", "Sentence $SentNum");
	
	print "\n            Processing <sentence ID=\"$SentenceNode->{ID}\"> in ";
	print "$Document P$Paragraph, S$SentNum\n";
	$ErrorFlag = 0;

    # annotationSet introduces a new frame. Create a node for it.

    } elsif ($elem eq 'annotationSet') {
	if (exists($attr{frameName})) {
	    ValidateDG($SentenceNode);
	    snapshot($SentenceNode);	# Create a dot file of the current state.
	    $FrameNode = MakeSimpleNode('Frame');
	    $FrameNode->{frameName} = $attr{frameName};
	    if ($Debug > 0) {
		print STDERR "Frame: $attr{frameName}\n";
	    }
	} else {
	    $FrameNode = undef;
	}
	
    # Store layer name

    } elsif ($elem eq 'layer') {
	$LayerName = $attr{name};
	
    # Reset the governor values

    } elsif ($elem eq 'labels') {
	if (defined($FrameNode)) {
	    $XNode = undef;
	    $GovNode = undef;
	}

    # Labels are processed here.

    } elsif ($elem eq 'label') {
	if (defined($FrameNode)) {
	    $LabelName = $attr{name};
	    
	    # NOT a null instantiation
	    if (exists($attr{start}) && exists($attr{end})) {
		
		# Target label
		
		if ($LabelName eq 'Target' && $LayerName eq 'Target') {
		    MakeOrAddTarget($LabelName, $attr{start}, $attr{end});
		    
		    # Non-target label
		} elsif ($LayerName eq 'FE' ||
			 $LayerName eq 'GF' ||
			 $LayerName eq 'PT' ||
			 $LabelName eq 'Null') {
		    MakeOrAddLabel($LabelName, $attr{start}, $attr{end});
		    if ($LabelName eq 'Null') {
			$LayerName= 'PT';
			MakeOrAddLabel('NP', $attr{start}, $attr{end});
			$LayerName='Other';
		    }
		    
		    # If it's not a Target, Supp, Cop, FE, GF, or PT, then it must be an Other, Sent, Noun, etc.

		} elsif ($LabelName eq 'Gov' ||
			 $LabelName eq 'Ctrlr' ||
			 $LabelName eq 'Supp' ||
			 $LabelName eq 'Cop' ) {
		    MakeOrAddGov($LabelName, $attr{start}, $attr{end});

		} elsif ($LabelName eq 'Ant') {

		    MakeOrAddAnt($LabelName, $attr{start}, $attr{end});
		    
		    # Handle "X" (range for Gov)
		} elsif ($LabelName eq 'X') {
		    $XNode=GetOrCreateNode($attr{start}, $attr{end});
		} 

	    # The following bit, if uncommented, handles apositional tags as
	    # annotations on the FrameNode.

	    } elsif (exists($attr{itype})) {
#		$FrameNode->{$LayerName}={} unless (exists($FrameNode->{$LayerName}));
#		$FrameNode->{$LayerName}{$attr{itype}}=[] unless (exists($FrameNode->{$LayerName}{$attr{itype}}));
#		push(@{$FrameNode->{LayerName}{$attr{itype}}}, $LabelName);
	    }

	# Named Entity labels are included by introducing frames with one denoted FE
	# whose Targets are the labeled words.  I.e., the NER label acts as Target label
	# and FE label in one.

	} elsif ($LayerName eq 'NER' && exists($attr{start}) && exists($attr{end})) {
	    $LabelName=$attr{name};

#	    print "Found NamedEntity NE:$LabelName\n";
	    $FrameNode=MakeSimpleNode('Frame');
	    $FrameNode->{frameName}="NE:$LabelName";
	    $FrameNode->{type}='Frame';
	    $FrameNode->{denoted}=[$LabelName];
	    MakeOrAddTarget('Target', $attr{start}, $attr{end});
	    PositionFrameAndTargets();
	    $FrameNode=undef;
	}
    }
} # FTStartTag()

######################################################################
#
# Handle close tags. In general, I try to invalidate variables that
# should only be valid within tags. This will help find invalid
# XML and bugs in the parser.
#

sub FTEndTag {
    my($expat, $elem) = @_;

    if ($elem eq 'text') {
	$SentenceNode->{text}=$Text;
	FTEndTagText($expat);
	AddPOS($SentenceNode);

    } elsif ($elem eq 'annotationSet') {

	if (exists($FrameNode->{frameName}) && $FrameNode->{frameName} !~ /^NE/) {
	    PositionFrameAndTargets();
	    if (!$TargetError &&
		exists($FrameNode->{governingNodes}) && $#{$FrameNode->{governingNodes}}>=0 ||
		exists($FrameNode->{ants}) && $#{$FrameNode->{ants}}>=0) {
		ProcessGovs();
	    }
	}
	$TargetError=0;
	$FrameNode = undef;
	%TargetType= ();

	# Delete any created frame elements that don't have a PT and a GF
    } elsif ($elem eq 'layers') {
	DeleteEmptyFE();

    } elsif ($elem eq 'layer') {
	$LayerName = undef;

	# When we see </labels>, reset any "Gov" that had an "X"
    } elsif ($elem eq 'labels') {
	if (defined($XNode) && defined($GovNode)) {
	    ProcessGovX();
	} 

	$XNode = undef;
	$GovNode = undef;
	

    } elsif ($elem eq 'label') {
	$LabelName = undef;

	# Below is where the sentence is passed back to the caller.

    } elsif ($elem eq 'sentence') {

	ValidateDG($SentenceNode);
	snapshot($SentenceNode);
	PostProcessSentence();
	ValidateDG($SentenceNode);
	snapshot($SentenceNode);

	# Call the user-provided callback with the sentence.

	&$Callback($SentenceNode, $ErrorFlag);
	$SentNum++;
    }	
} # FTEndTag()

#########################################################
#
# This routine looks at the various Govs of the current
# frame and ranks them by the POS of the headword (V>Prep>N).
# The higher ranked ones make a non-local child edge to the 
# next lowest, on the assumption that a Support Verb will 
# take a Support Prep (if present) as its direct complement.   
# If two Supps have the same rank, they are both put in as 
# parents of the next lower down.  In any case, Antecedents are
# at the top of the chain (i.e. not embedded under anything) and the
# FrameNode itself is at the bottom of this chain.
#
# Unlike other nlEdges, these can easily be wrong, as inter-
# vening material should quite possibly be included in the 
# childNode of a Supp:
#    '[take Supp] a [bath T]', which should be '[take [a bath NP]]').
#                 ^
# In PostProcessMoveGoverned, the nlEdges that didn't appear to work
# will be reexamined, although this hasn't been implemented.
#

sub ProcessGovs {

    if ($TargetError) {
	foreach my $gov (@{$FrameNode->{governingNodes}}) {
	    RemoveFromList($gov->{governedNodes}, $FrameNode);
	}
	return;
    }
    
    $FrameNode->{ants}=[] unless (exists($FrameNode->{ants}));
    $FrameNode->{governingNodes}=[] unless (exists($FrameNode->{governingNodes}));
    
    my @governorChain=@{$FrameNode->{governingNodes}};
    @governorChain=sort { $TypeRanking{$PTtoType{$HeadPOStoPTs{$a->{govWord}{POS}}[0]}} <=> 
			      $TypeRanking{$PTtoType{$HeadPOStoPTs{$b->{govWord}{POS}}[0]}} } @governorChain; 

    my @antecedents=@{$FrameNode->{ants}};    
    @governorChain=(@antecedents, @governorChain);
    push(@governorChain, ($FrameNode));
    
    my $n=0; 
    while($n<$#governorChain) {

	my $m=1;

	# This whileloop loops through until we're either right before the FrameNode, 
	# or the Rank stops being the same.  Same-rank govs should have the same parents,
	# so this loop gets all the parents of the first and imposes them on the rest.

	while ($n+$m<$#governorChain && 
	       ($governorChain[$n]{type} eq 'List' && $governorChain[$n+$m]{type} eq 'List' ||
	       ($governorChain[$n]{type} ne 'List' && $governorChain[$n+$m]{type} ne 'List' &&
	       $TypeRanking{$PTtoType{$HeadPOStoPTs{$governorChain[$n]{govWord}{POS}}[0]}} == 
	       $TypeRanking{$PTtoType{$HeadPOStoPTs{$governorChain[$n+$m]{govWord}{POS}}[0]}}))) {
	    if (exists($governorChain[$n]{nlParentEdges})) {
		foreach my $parent (map {$_->{parentNode}} @{$governorChain[$n]{nlParentEdges}}) {
		    if (FindInList($parent, \@governorChain)) {
			MakeNLEdge($parent, $governorChain[$n+$m]);
		    }
		}
	    }
	    $m++;
	}

	# This forloop loops through however many Ants/Govs have the same rank (determined as $m in loop above)
	# and makes NL edges from each group to the next lower rank; this usually means NL edges to the 
	# FrameNode, which is at the bottom.

	for (my $k=0;$k<$m;$k++) {

#	    warn("parent gov: '", NodeToText($governorChain[$n+$k]), "', child gov: '", NodeToText($governorChain[$n+$m]), "'\n");

	    my $proposedEdge=MakeNLEdge($governorChain[$n+$k], $governorChain[$n+$m]);
	    
	    if ($governorChain[$n+$k]{type} eq 'List') {
		$proposedEdge->{type}='Relativized';	    
	    } elsif ($governorChain[$n+$k]{type} eq 'Supp' || 
		     $governorChain[$n+$k]{type} eq 'Cop') {
		$proposedEdge->{type}='SemHead';
	    }
	}
	$n=$n+$m;
    } 
}

######################################################################
#
# Any FE that doesn't have a PT and a GF should just be deleted.
#

sub DeleteEmptyFE {
    my($nledge, $child);
    foreach $nledge (@{$FrameNode->{nlChildEdges}}) {
	if (exists($nledge->{FE}) && (!exists($nledge->{PT}) || !exists($nledge->{GF})) 
#	    ||
#	    exists($nledge->{PT}) && !exists($nledge->{FE}) ||
#	    exists($nledge->{GF}) && !exists($nledge->{FE})
	    ) {
	    RemoveFromList($nledge->{parentNode}{nlChildEdges}, $nledge);
	    RemoveFromList($nledge->{childNode}{nlParentEdges}, $nledge);
	}
    }
} # DeleteEmptyFE()

######################################################################
#
# Called to reset the destination span of a Gov.
# Only called when an X appeared.
#

sub ProcessGovX {

    my $nlEdge=MakeNLEdge($GovNode, $XNode);
    if ($GovNode->{type} eq 'Supp' || $GovNode->{type} eq 'Cop') {
	$nlEdge->{type}='SemHead';
    } elsif ($GovNode->{type} eq 'Ant') {
	$nlEdge->{PT}='Srel';
    } 
    RemoveFromList($FrameNode->{governingNodes}, $GovNode);

} # ProcessGovX()


######################################################################
#
# Handle the text of the sentence. Splits it into space delimited 
# words, created a word node for each. Also does a bunch of
# initialization on global variables used in the parse.
#

sub FTEndTagText {
    my($expat) = @_;
    my($wordform, $start, $end);
    $SentenceNode->{start} = 0;
    $SentenceNode->{end} = length($Text)-1;
    $start = 0;
    while (1) {
	$end = index($Text, ' ', $start)-1;
	if ($end <= -1) {
	    $end = length($Text)-1;
	}
	
	$wordform = substr($Text, $start, $end-$start+1);

	# The commented method below is probably superior, since it doesn't
	# require that the wordnodes come first.  If they come at some other
	# point, they are added under the lowest appropriate node. But it is
	# longer, and we don't need it at the moment.

#	my $parentNode=$SentenceNode;
#	while (exists($parentNode->{childEdges}) && 
#	       $parentNode->{start} <= $start && 
#	       $parentNode->{end} >= $end){
#	    my $n=0;
#	    while ($parentNode->{childEdges}[$n]{childNode}{start} <= $start) {
#		$n++;
#	    }
#	    $parentNode=$parentNode->{childEdges}[$n-1]{childNode};
#	}
#	MakeWordNode($parentNode, $wordform, $start, $end);
	
	MakeWordNode($SentenceNode, $wordform, $start, $end);
	
	last if ($end >= length($Text)-1);
	$start = $end+2;	    
    }
    $InText = 0;
} # FTEndTagText()

sub MakeWordNode {
    my($parentNode, $wordform, $start, $end) = @_;
    my($node, $edge);
    $node = MakeSimpleNode('Word');
    $node->{start} = $start;
    $node->{end} = $end;
    $node->{form} = $wordform;
    $StartToWord{$start} = $node;
    $EndToWord{$end} = $node;
    $edge = MakeSimpleEdge($parentNode, $node);
    return $node;
} # MakeWordNode()


sub FTText {
    my($expat, $string) = @_;
    if ($InText) {
	$Text .= $string;
    }
} # FTText()

######################################################################
#
# Make a target node of a frame. Assumes the FrameNode already
# exists. 

sub MakeOrAddTarget {

    my($type, $start, $end) = @_;
    my($targetnode, @targetnodes);

    # GetOrCr. would just fetch the wordnode that has already been made unless a single label covers two targets.
#    $targetnode = GetOrCreateNode($start, $end);

    # The new method finds all the word nodes within the start-end and identifies each of these as targets.
    # NB: the list of all words is obtained with: values %StartToWord

    @targetnodes=grep {$_->{start} >= $start && $_->{end} <= $end} values %StartToWord;

    push(@{$FrameNode->{targetNodes}}, @targetnodes);
    foreach $targetnode (@targetnodes) {
	$TargetType{$targetnode->{sym}}=$type;
	my $edgetype=$targetnode->{parentEdge}{type};
	$TargetEdgeType{$targetnode->{sym}}=$edgetype;
	if ($edgetype eq 'Gov' ||
	    $edgetype eq 'Supp' ||
	    $edgetype eq 'Cop' ||
	    $edgetype eq 'Ctrlr') {

	    warn("Deleting Gov $targetnode->{parentEdge}{parentNode}{sym}, to be replaced by frame for target $targetnode->{sym}\n");
	    $targetnode->{parentEdge}{parentNode}{type}='List';
	    delete $targetnode->{parentEdge}{parentNode}{governedNodes};
	    delete $targetnode->{parentEdge}{parentNode}{nlChildEdges};
	    foreach my $childedge (@{$targetnode->{parentEdge}{parentNode}{childEdges}}, ) {
		$childedge->{type}='NonHead';
	    }
	    $TargetEdgeType{$targetnode->{sym}}='NonHead';		

	} elsif ($edgetype eq 'Ant') {
	    warn("Target $targetnode->{form} and frame $FrameNode->{frameName} being placed under Antecedent node $targetnode->{parentEdge}{parentNode}{sym}:\n     ", NodeToText($targetnode->{parentEdge}{parentNode}), "\nHopefully this is correct...\n");
	} elsif ($edgetype eq 'Target') {
	    if ($targetnode->{parentEdge}{parentNode}{frameName} !~ /^NE/ ) {
		error("$targetnode->{sym}, '", NodeToText($targetnode), "', already a target of node $targetnode->{parentEdge}{parentNode}{sym}\n");
		error("Can't delete Frame $targetnode->{parentEdge}{parentNode}{frameName} on $targetnode->{parentEdge}{parentNode}{sym}\n");
		$TargetError=1;
#	        $targetnode->{parentEdge}=$targetnode->{parentEdge}{parentNode}{parentEdge};
#	        $targetnode->{parentEdge}{childNode}=$targetnode;	    
	    } else {
		warn("Removing clashing NE annotation on '", NodeToText($targetnode), "'.\n");
		$targetnode->{parentEdge}{parentNode}{type}='List';
		delete $targetnode->{parentEdge}{parentNode}{frameName};
		delete $targetnode->{parentEdge}{parentNode}{targetNodes};
		delete $targetnode->{parentEdge}{parentNode}{denoted};
		foreach my $childedge (@{$targetnode->{parentEdge}{parentNode}{childEdges}}) {
		    $childedge->{type}='NonHead';
		}
		$TargetEdgeType{$targetnode->{sym}}='NonHead';		
	    }
	}
    }
}  # MakeOrAddTarget()

###################################################################
# 
# When finished with the AnnoSet, put the frame node under whereever the
# target(s) was/were originally, and put the target(s) and anything between them under it.
#
# If a target node already exists, the parent of the new target must
# be the same as the parent of the frame.
#


sub PositionFrameAndTargets {

    return if (!defined($FrameNode));

    if (!exists($FrameNode->{frameName}) || !defined($FrameNode->{frameName})) {
	warn "Ignoring non-frame annotation.\n";
	$TargetError=1;
    }
    
    if ($TargetError ||
	!exists($FrameNode->{targetNodes}) || 
	!defined($FrameNode->{targetNodes}) ||
	$#{$FrameNode->{targetNodes}} < 0 ||
	!defined($FrameNode->{targetNodes}[0])) {

	warn("No target defined; ignoring annotation for frame $FrameNode->{frameName}.\n") if (!$TargetError);
	foreach my $child (@{$FrameNode->{nlChildEdges}}) {
	    RemoveFromList($child->{childNode}{nlParentEdges}, $child)
	}

#	Eventually, this will accept non-target-based annotation sets; $start, $end would be defined
#	as $start, $end of the nlChildNodes? or a particular nlChildNode?
	
	return;
    } 

    # This handles multi-Target labels

    my($start, $end);


    $start = $FrameNode->{targetNodes}[0]{start};
    $end = $FrameNode->{targetNodes}[0]{end};
    foreach my $targetnode (@{$FrameNode->{targetNodes}}) {
	
	if ($start >= $targetnode->{start}) {
	    $start = $targetnode->{start};
	} elsif ($end <= $targetnode->{end}) {
	    $end=$targetnode->{end};
	}
	    
	# Otherwise, target is inside target of same frame and annotation set.  Only possible 
	# if $#{$FrameNode->{targetNodes}} > 1, which is rare, but unremarkable.
	    
    }

    my $pnode=$FrameNode->{targetNodes}[0]{parentEdge}{parentNode};

    my $tempFrameNode = CreateNewNode($start, $end, $pnode);

    $FrameNode->{childEdges}=$tempFrameNode->{childEdges};
    $FrameNode->{start}=$tempFrameNode->{start};
    $FrameNode->{end}=$tempFrameNode->{end};
    $FrameNode->{sym}=$tempFrameNode->{sym};

    my $pedge = $tempFrameNode->{parentEdge};
    $pedge->{type}=$TargetEdgeType{$FrameNode->{targetNodes}[0]{sym}};
    $pedge->{childNode}=$FrameNode;
    $FrameNode->{parentEdge}=$pedge;    
 
    foreach my $targetnode (@{$FrameNode->{targetNodes}}) {
	$targetnode->{parentEdge}{type}=$TargetType{$targetnode->{sym}};
    }

    # Here we take all the non-local edges that have become local and delete them from nlEdges, as well as deleting nlParent from child.

    my @nlEdges=@{$FrameNode->{nlChildEdges}};
    foreach my $localEdge (@{$FrameNode->{childEdges}}) {
	$localEdge->{parentNode}=$FrameNode;
	for (my $i = 0; $i <= $#nlEdges; $i++) {
	    if ($nlEdges[$i]{childNode} eq $localEdge->{childNode}) {

		RemoveFromList($localEdge->{childNode}{nlParentEdges}, $nlEdges[$i]);
		splice(@nlEdges, $i, 1);
		last;
	    }
	}
    }
}

######################################################################
#
# Given a node, reset its parent to the given parent node. Does
# not touch non-local links. No edges are created or destroyed;
# they only have their parentNode changed.
#
# NOTE: Need to update start/end of ALL parents of old and new parent.

sub Reparent {
    my($node, $newparentnode) = @_;
    my($childedges, $i, $oldparentnode);

    # Find the list of child edges out of the current parent
    # of $node.

    $oldparentnode = $node->{parentEdge}{parentNode};
    $childedges = $oldparentnode->{childEdges};

    # Now delete the parent edge ref of the $node from that list
    for ($i = 0; $i <= $#{$childedges}; $i++) {
	if ($childedges->[$i] eq $node->{parentEdge}) {
	    splice(@{$childedges}, $i, 1);
	    last;
	}
    }

    # Set the start/end of the old parent node (in case the removed
    # child was the first or last).

    UpdateStartEnd($oldparentnode);
	
    # Set the node's parent correctly.
    $node->{parentEdge}{parentNode} = $newparentnode;

    InsertChildEdge($newparentnode, $node->{parentEdge});
} # Reparent()


######################################################################
#
# NOTE: This will add the edge to the childEdges in the parentNode.
# It is an error if the childNode already has a parent.
#

sub MakeSimpleEdge {
    my($parentNode, $childNode) = @_;
    my($edge);
    $edge = { metatype => 'Edge',
	      type => 'NonHead',
	      parentNode => $parentNode,
	      childNode => $childNode,
	      sym => sprintf("e%03d", $SymIndex++)
	      };
    if (exists($childNode->{parentEdge}) && defined($childNode->{parentEdge})) {
	error("MakeSimpleEdge passed with a child that already has a parent.");
    }
    $childNode->{parentEdge} = $edge;

    InsertChildEdge($parentNode, $edge);
    $SymToRef{$edge->{sym}} = $edge;
    return $edge;
} # MakeSimpleEdge()

######################################################################
#
#
#

sub MakeNLEdge {
    my($parentNode, $childNode) = @_;
    my($edge);
    $edge = { metatype => 'nlEdge',
	      type => 'NonHead',
	      parentNode => $parentNode,
	      childNode => $childNode,
	      sym => sprintf("e%03d", $SymIndex++)
	      };
    InsertEdge($parentNode->{nlChildEdges}, $edge);
    InsertEdge($childNode->{nlParentEdges}, $edge);
    $SymToRef{$edge->{sym}} = $edge;
} # MakeNLEdge()

######################################################################
#
# Insert the given edge into the child edge list of the node in the
# correct place (i.e. so the list remains sorted by start position).
# Update start/end of the $node from the new edge's start/end (and
# all ancestors).
#

sub InsertChildEdge {
    my($node, $newedge) = @_;

    InsertEdge($node->{childEdges}, $newedge);
    UpdateStartEnd($node);

} # InsertChildEdge();

sub UpdateStartEnd {
    my($node) = @_;
    my($childedges);

    while (1) {
	$childedges = $node->{childEdges};
	if ($#{$childedges} >= 0) {
	    $node->{start} = $childedges->[0]{childNode}{start};
	    $node->{end} = $childedges->[$#{$childedges}]{childNode}{end};
	} else {
	    delete($node->{start});
	    delete($node->{end});
	}
	last if (!exists($node->{parentEdge}) || !defined($node->{parentEdge}));
	$node = $node->{parentEdge}{parentNode};
    }
} # UpdateStartEnd()

######################################################################
#
# Insert the newedge into the given edge list in the correct place.
# Does NOT update start/end (since that's not meaningful for nl).
#

sub InsertEdge {
    my($edges, $newedge) = @_;
    my($i);

    # Figure out where to insert the new edge into childEdges.
    for ($i = 0; $i <= $#{$edges}; $i++) {
	if (!exists($newedge->{childNode}) || 
	    !exists($newedge->{childNode}{start}) || 
	    !defined($newedge->{childNode}{start})) {
	    error ("New edge $newedge->{parentNode}{sym} to $newedge->{childNode}{sym} not defined\n");
	    return;
	}
	last if ($newedge->{childNode}{start} < $edges->[$i]{childNode}{start});
    }
    splice(@{$edges}, $i, 0, $newedge);
} # InsertEdge()


######################################################################
#
# Create a new empty node.
#

sub MakeSimpleNode {
    my($type) = @_;
    my($node);
    $node = { type => $type,
	      childEdges => [],
	      nlChildEdges => [],
	      nlParentEdges => [],
	      sym => sprintf("n%03d", $SymIndex++)
	  };
    push(@AllNodes, $node);
    $SymToRef{$node->{sym}} = $node;
    return $node;
} # MakeSimpleNode()

######################################################################
#
# Given a label, either add the name to an existing non-local edge of
# the current frame, or create a non-local edge to the (possible new)
# node.
#

sub MakeOrAddLabel {
    my($name, $start, $end) = @_;
    my($edge, $node, $targetnode);


    # Check if the label overlaps any targets. If so, add it to "denoted" list.
    foreach $targetnode (@{$FrameNode->{targetNodes}}) {
	if ($start <= $targetnode->{end} &&
	    $end >= $targetnode->{start}) {
	    push(@{$FrameNode->{denoted}}, $name);
	    return;
	}
    }

    # Check if an edge to the desired node already exists
    foreach $edge (@{$FrameNode->{nlChildEdges}}, @{$FrameNode->{childEdges}}) {
	if ($edge->{childNode} eq GetOrCreateNode($start, $end)) {
	    $edge->{$LayerName} = $name;
	    return;
	}
    }

    # If we make it here, the edge doesn't exist.

    $node = GetOrCreateNode($start, $end);
    $edge = MakeNLEdge($FrameNode, $node);
    $edge->{$LayerName} = $name; 
    $edge->{type} = 'Dep';
} # MakeOrAddLabel()

######################################################################
#
# 
#

sub MakeOrAddGov {
    my($name, $start, $end) = @_;
    my($node, $pedge, $edge);
    my($govNode);

    # The following takes any label that might be present, finds the individual words that cover that
    # span, and makes a separate node for each.  The dependencies between these nodes and the FrameNode
    # are figured out in ProcessGovs.

    my @govWords=grep {$_->{start} >= $start && $_->{end} <= $end} values %StartToWord;
    foreach my $govWord (@govWords) {
	next if (exists($govWord->{parentEdge}) && 
		 exists($govWord->{parentEdge}{type}) && 
		 $govWord->{parentEdge}{type} eq 'Target');

	$govNode = MakeSimpleNode($name);
	$govNode->{start} = $govWord->{start};
	$govNode->{end} = $govWord->{end};
	$pedge = $govWord->{parentEdge};
	$pedge->{childNode} = $govNode;
	$govNode->{parentEdge} = $pedge;
	delete($govWord->{parentEdge});
	$edge = MakeSimpleEdge($govNode, $govWord);
	$edge->{type} = $name;
	$govNode->{govWord}=$govWord;
	$govNode->{PT}=$govWord->{PT} if (exists($govWord->{PT}));

	if (!defined($GovNode) || $name eq 'Gov') {
	    $GovNode=$govNode;
	}

    # The following additional features on govNodes and FrameNodes enable post-processing to put Frames under govs.

	$govNode->{governedNodes}=[] unless (exists($govNode->{governedNodes}));
	push (@{$govNode->{governedNodes}}, $FrameNode);
	$FrameNode->{governingNodes}=[] unless exists($FrameNode->{governingNodes});
	push (@{$FrameNode->{governingNodes}}, $govNode);
    }

} # MakeOrAddGov()

###############################################
# 
# Dealing with an Ant label requires making a 
# label on the Other layer, as well as making an
# extra phrase that is supposed to contain the Ant
# and the FrameNode.

sub MakeOrAddAnt {
    my($name, $start, $end)=@_;


    MakeOrAddLabel($name, $start, $end);
    my $node=GetOrCreateNode($start, $end);

    # Make List node and insert it at the position of
    # the Ant $node.

    my $antNode = MakeSimpleNode('List');
    $antNode->{start} = $start;
    $antNode->{end} = $end;
    my $pedge = $node->{parentEdge};
    $pedge->{childNode} = $antNode;
    $antNode->{parentEdge} = $pedge;
    delete($node->{parentEdge});
    my $edge = MakeSimpleEdge($antNode, $node);
    $edge->{type} = 'Head';
    $antNode->{PT}=$node->{PT} if (exists($node->{PT}));

    $FrameNode->{ants}=[] unless (exists($FrameNode->{ants}));
    push(@{$FrameNode->{ants}}, ($antNode));
}


######################################################################
#
# Removes the given element from the given list reference.
# Return 1 if it was there, 0 otherwise.
#

sub RemoveFromList {
    my($ls, $elem) = @_;
    my($ii);
    for ($ii = 0; $ii <= $#{$ls}; $ii++) {
	if ($ls->[$ii] eq $elem) {
	    splice(@{$ls}, $ii, 1);
	    return 1;
	}
    }
    warn("Couldn't find element in list: ");
    dmsg(Dumper($ls), "Sought element: ", Dumper $elem);
    return 0;
} # RemoveFromList()

######################################################################
#
# Given a start and end position, return a node that contains all
# words in the range. 
# 

# If no such node exists, create an empty phrase node containing all
# nodes within the range.
#

sub GetOrCreateNode {
    my($start, $end) = @_;
    my($node, $parent);

    # Check for exact match. This could be optimized with a hash.

    foreach $node (@AllNodes) {
	if (exists($node->{start}) && $node->{start} == $start && 
	    exists($node->{end}) && $node->{end} == $end) {

	    # Return the highest node with the given span

	    while (exists($node->{parentEdge}) &&
		   $node->{parentEdge}{parentNode}{start} == $start &&
		   $node->{parentEdge}{parentNode}{end} == $end) {
		$node = $node->{parentEdge}{parentNode};
	    }
	    return $node;
	}
    }



    # If the parent is
    # a simple phrase (non-frame), then 
    #
    # If the parent is a frame, just return the parent, since you 
    # can't steal targets from frames.

    $parent=FindContainingNode($start, $end);

    if (exists($parent->{frameName})) {
	return $parent;
    }

	# If we get here, we want to create a new node containing the span.
        # But first we have to test for bracketing problems.
    
    my($startFound, $endFound);
    foreach my $subParent (map {$_->{childNode}} @{$parent->{childEdges}}) {
	if ($subParent->{start} == $start) {
	    $startFound=1;
	}
	if ($subParent->{end} == $end) {
	    $endFound=1;
	}
    }

    if ($startFound && $endFound) {		
	return CreateNewNode($start, $end, $parent);
    } 

    
    if (!$startFound) {
	error("Bad bracketing when trying to make a label of frame $FrameNode->{frameName} from $start to $end: \n", substr($SentenceNode->{text}, $start, $end-$start+1));
	foreach my $subParent (sort {$b->{start} <=> $a->{start}} map {$_->{childNode}} @{$parent->{childEdges}}) {
	    if ($subParent->{start} < $start) {
		print "Bracketing problem appears to involve node $subParent->{sym}: \n", NodeToText($subParent), "\n\n";
		return $subParent;
	    }
	} 
    } elsif (!$endFound) {
	error("Bad bracketing when trying to make a label of frame $FrameNode->{frameName} from $start to $end: \n", substr($SentenceNode->{text}, $start, $end-$start+1));
	foreach my $subParent (sort {$a->{end} <=> $b->{end}} map {$_->{childNode}} @{$parent->{childEdges}}) {
	    if ($subParent->{end} > $end) {
		print "Bracketing problem appears to involve node $subParent->{sym}: \n", NodeToText($subParent), "\n\n";
		return $subParent;
	    }
	}
    }
} # GetOrCreateNode()


###############################################################
#
# Return the lowest node that has a span containing $start to $end.
# Start with the word that starts at $start, and move to the
# word's parent until the parents's end >= $end. 

sub FindContainingNode {
    my($start, $end)=@_;

    my $parent;
    if (exists($StartToWord{$start})) {
	$parent = $StartToWord{$start}; 
	while ($parent->{end} < $end) {
	    if (!defined($parent->{parentEdge}{parentNode})) {
		error('node with no parent');
		die "Hit badness at $parent->{sym}; yuck!\n";
	    }
	    $parent = $parent->{parentEdge}{parentNode};
	}
	return $parent;
    } else {
	error('Word nodes not defined; <text> must precede <annotationSets> in fulltext XML, and tags must align with words.');
	error('Node from character position $start to $end does not line up with any word.');
#	my @parents=FindNode($SentenceNode, sub {$_[0]->{start}});
    }

} # End FindContainingNode

###############################################################
# Extract the children of parent
# in the span $start/$end, and reparent to a newly created node.
# Make the newly created node a child of the parent.

sub CreateNewNode {

    my($start, $end, $parent)=@_;
    my($node, $child, @children);
    
    # If we get here, the parent is a simple phrase.

    $node = MakeSimpleNode('List');

    foreach $child (map { $_->{childNode} } @{$parent->{childEdges}}) {
	if ($child->{start} > $end) {
	    last;
	}
	if ($child->{start} >= $start) {
	    Reparent($child, $node);
	}
    }
    MakeSimpleEdge($parent, $node);
    return $node;
} 

######################################################################
#
# Check if the graph appears to be valid.
#
# Currently, the following checks are implemented:
#
# * Each of my children has parent equal to me.
# * My children are contiguous.
# * My children are valid.
# * node's start = first child's start.
# * node's end   = last  child's end.
# * I am in the list of my non-local children's non-local parents.

sub ValidateDG {
    my($node) = @_;
    my($childedge, $childnode, $nchildren, $i);

    $nchildren = $#{$node->{childEdges}}+1;
	
    for ($i = 0; $i < $nchildren; $i++) {
	$childedge = $node->{childEdges}[$i];
	$childnode = $childedge->{childNode};

	# Check parent/child relationship
	if ($childedge->{parentNode} ne $node) {
	    error("Parent of edge $childedge->{sym} is not $node->{sym}");
	}
	if ($childnode->{parentEdge} ne $childedge) {
	    error("Parent of node $childnode->{sym} is not $childedge->{sym}");
	}

	# Check continuity
	if ($i < $nchildren-1 && 
	    (($childnode->{end}+2) !=  $node->{childEdges}[$i+1]{childNode}{start})) {
	    error("Child $i and " . ($i + 1) . " of node $node->{sym} are not contiguous");
	}
	
	# Check that my children are valid.
	ValidateDG($childnode);
    }

    # Check that my start/end match my first/last child's start/end
    if ($nchildren > 0) {
	if ($node->{start} != $node->{childEdges}[0]->{childNode}{start}) {
	    error("Start of $node->{sym} is not equal to first child's start");
	}
	if ($node->{end} != $node->{childEdges}[$nchildren-1]->{childNode}{end}) {
	    error("End of $node->{sym} is not equal to last child's end");
	}
    }

    # Check that non-local children list me as a non-local parent

    foreach $childedge (@{$node->{nlChildEdges}}) {
	$childnode = $childedge->{childNode};
	if ($childedge->{parentNode} ne $node) {
	    error("Parent of non-local edge $childedge->{sym} is not $node->{sym}");
	}
	# Check that $node is on the list of non-local parents.
	if ((grep { $_ eq $childedge } @{$childnode->{nlParentEdges}}) != 1) {
	    error("$node->{sym} is either not listed or listed more than once in non-local children of $childnode->{sym}");
	}
    }
} # ValidateDG()

sub PostProcessSentence {

    my @unmovableGFs=('Ext', 'Appositive', 'Head');


    PostProcessMoveUp();
    PostProcessRemoveSingletonGovs();
    AssignNodePTs();

    snapshot($SentenceNode);

    while (PostProcessMoveLocals(\@unmovableGFs)) { }

    snapshot($SentenceNode);

    PostProcessRemoveSingletonLists();

    ReverseModDependencies();

    PostProcessGovs();
    PostProcessRemoveSingletonLists();

    ReverseModDependencies();
    @unmovableGFs=('Appositive', 'Head');
    while (PostProcessMoveLocals(\@unmovableGFs)) { }
##    while (PostProcessMoveGoverned(\@unmovableGFs)) { }
    PostProcessRemoveSingletonLists();
    FindHeads($SentenceNode);

} # PostProcessSentence()



##################################################################
#
# Probably replaceable by grep {$_ eq $item} @{$list}???
#
#

sub FindInList {
    my ($item, $list)=@_;

# This (commented out) code is here simply because &FindInList
# is called in the parts of the code that most often go wrong.
# It is thus a good place to see what is going on, without putting
# prints after every line of code or relying on the perldebugger.
#
#    if ($ErrorFlag) {
#	$Data::Dumper::Maxdepth=3;
#	print 'Item:  ', Dumper($item);
#	print 'List:  ', Dumper($list);
#    }

    foreach my $member (@{$list}) {
	if ($member eq $item) {
	    return 1;
	}
    }
    return 0;
}

##################################################################

sub AssignNodePTs {

    foreach my $ptNode (FindNodes($SentenceNode, sub {!exists($_[0]->{PT}) || !defined($_[0]->{PT})})) { 
	my @ptList;
	foreach my $parentEdge (@{$ptNode->{nlParentEdges}}, ($ptNode->{parentEdge})) {
	    if (exists($parentEdge->{PT}) && !FindInList($parentEdge->{PT}, \@ptList)) {
		push(@ptList, ($parentEdge->{PT}));
	    }
	}
	if ($#ptList>=0) {
	    if ($#ptList>0) {
		error("Inconsistent PTs, (", join(' vs. ', @ptList), ") for following span:\n     ", NodeToText($ptNode));
	    }
	    $ptNode->{PT}=$ptList[0];
	} else {
	    foreach my $childEdge (@{$ptNode->{nlChildEdges}}, @{$ptNode->{childEdges}}) {
		if (exists($childEdge->{reversedEdge}) && exists($childEdge->{reversedEdge}{PT}) &&
		    !FindInList($childEdge->{reversedEdge}{PT}, \@ptList)) {
		    push(@ptList, ($childEdge->{reversedEdge}{PT}));
		}
	    }
	    if ($#ptList>=0) {
		if ($#ptList>0) {
		    warn("Multiple PTs, (",join(' vs. ', @ptList), ") for following span:\n     ", NodeToText($ptNode));
		}
		$ptNode->{PT}=$ptList[0];
	    } else {
#		warn("No PT found for following span:\n     ", NodeToText($ptNode));
	    }
	}
    }
}


#################################################################################################
# 
# This routine finds all edges that have GF Head and finds or creates an edge with reversed
# child and parent.
#

sub ReverseModDependencies {

    foreach my $edge (FindEdges($SentenceNode, sub 
				{(exists($_[0]->{GF}) && defined($_[0]->{GF}) && $_[0]->{GF} eq 'Head') || 
				     (
				     (exists($_[0]->{Other}) && defined($_[0]->{Other}) && $_[0]->{Other} eq 'Ant') ||
				      (exists($_[0]->{parentNode}{parentEdge}) && 
				       defined($_[0]->{parentNode}{parentEdge}) &&
				       exists($_[0]->{parentNode}{parentEdge}{parentNode}) &&
				       defined($_[0]->{parentNode}{parentEdge}{parentNode}) &&
				       exists($_[0]->{childNode}) && 
				       defined($_[0]->{childNode}) && 
				       $_[0]->{parentNode}{parentEdge}{parentNode} eq $_[0]->{childNode}))})
		      ) {
		      

	my $reversedEdge;
	
	my @revEdges=FindEdges($edge->{childNode}, sub {$_[0]->{childNode} eq $_[1] &&
							  $_[0]->{parentNode} eq $_[2]}, 
			       $edge->{parentNode}, $edge->{childNode});
	if ($#revEdges>=0) {
	    if ($#revEdges>0) {
		warn "More than one edge between $edge->{childNode}{sym} and $edge->{parentNode}{sym}\nChoosing one, perhaps wrongly.\n";
	    }
	    $reversedEdge=$revEdges[0];

	# if no reversed edge is found, make a new one

	} else {

	    # if the new parentNode isn't a word, life is easy...
	    # if the new parent node is the syntactic head of its phrase, that's also easy... 
	    #
	    # All of these special cases are now taken care of in MoveLocal, so here we 
	    # just go ahead and make the edge, even if the parent is a word.

	    $reversedEdge = MakeNLEdge($edge->{childNode}, $edge->{parentNode});
	}

	$reversedEdge->{reversedEdge}=$edge;
	RemoveFromList($edge->{parentNode}{nlChildEdges}, $edge);
	RemoveFromList($edge->{childNode}{nlParentEdges}, $edge);
	$edge->{parentNode}={};
	$edge->{childNode}={};
	$edge->{type}='Reversed';

    }

} # End ReverseModDependencies

###########################################################################
#
# This routine copies non-local frame element (etc) edges from a frame to a 
# Ctrlr, Cop, or Supp parent.  We do this because we know that frames share at 
# least some arguments with a Ctrlr, Cop, and Supp, so if we didn't find 
# something locally initially, this may be because it is actually a syntactic 
# argument of the Ctrlr, Cop, or Supp of the frame.  This argument will then be moved
# local, as usual, on a later cycle.  Importantly, we should only be copying
# real arguments of the Frame, not anything else, so we exclude NonHead edges
# of the frame which arise due to modifiers.
#

sub PostProcessGovs {

    my @govNodes=FindNodes($SentenceNode, sub {$_[0]->{type} eq 'Ctrlr' ||
						 $_[0]->{type} eq 'Cop' ||
						 $_[0]->{type} eq 'Supp' });
    foreach my $gov (@govNodes) {
	foreach my $child (@{$gov->{governedNodes}}) {
	    foreach my $nlGrandChild (@{$child->{nlChildEdges}}) {
		next if ($nlGrandChild->{type} eq 'NonHead');
		my @edges=FindEdges($gov, sub { $_[0]->{parentNode} eq $_[1] && 
						    $_[0]->{childNode} eq $_[2]}, 
				    $gov, $nlGrandChild->{childNode});
		my $edge;
		if ($#edges<0){
		    $edge=MakeNLEdge($gov, $nlGrandChild->{childNode});
		} else {
		    $edge=$edges[0];
		}
		$edge->{type}=$nlGrandChild->{type};
		$edge->{PT}=$nlGrandChild->{PT} if exists($nlGrandChild->{PT});
		$edge->{FE}=$nlGrandChild->{FE} if exists($nlGrandChild->{FE});
		$edge->{reversedEdge}=$nlGrandChild->{reversedEdge} if (exists($nlGrandChild->{reversedEdge}));
	    }
	}
    }
}

#############################################################

sub FindHeads {

    my($node)=@_;

    # First we have to recognize special structures like conjunctions (which have multiple Heads), 
    # and auxiliary verbs and markers (which are SynHeads but not SemHeads).  If we do any
    # of these special things, we return after recursing down the tree.

    if ($node->{type} eq 'List' && exists($node->{PT}) && exists($MarkerToBare{$node->{PT}})) {
	AddMarker($node);
    }
    
    # Note that &ProcessConjunction checks to see if its conjuncts are marker nodes;
    # if they are marker nodes, then &AddMarker is run on each, and IsMarkerNode will test false 
    # on the parent node of the conjunction (because the type will already be 'MarkerPhrase', which
    # means that no work needs to be done).
    
# Conjunctions currently not implemented.
    
#    my $alreadyDone=0;
#    $alreadyDone +=ProcessConjunction($node);
    
    if (exists($node->{childEdges})) {
	foreach my $child (map {$_->{childNode}} @{$node->{childEdges}}) {
	    FindHeads($child);
	}
    }
    
#    return if($alreadyDone);
    
    ProcessAux($node);
    
    # Second, we find the head(s) of each phrase, which is done differently for each type.
    
    if (exists($node->{type})) {
	
	if ($node->{type} eq 'Word') {
	    AddWordPT($node);
	    return;
	} elsif ($node->{type} eq 'Frame') {
	    FindFrameHead($node);
	} elsif ($node->{type} eq 'Gov' || $node->{type} eq 'Ctrlr' || 
		 $node->{type} eq 'Supp' || $node->{type} eq 'Cop') {
	    FindGovHead($node);
	} elsif ($node->{type} eq 'List') {
	    FindListHead($node);
	} elsif ($node->{type} eq 'MarkerPhrase') {
#	    print"\n";
	} else {
	    print "Unknown type on node $node->{sym}.\n";
	}
    } else {
	error("Node $node->{sym} does not have type information.\n");
    }
    
    if(!exists($node->{headNode})) {
	NarrowHeads($node);
    }
    
    # Third, we propagate the PTs that were possible for the headNode(s) up
    # to be the possiblePTs of the current node.
    
    my $noHead=0;
    if (exists($node->{headNode}) && defined($node->{headNode}) && ref($node->{headNode}) eq 'ARRAY') {
        foreach my $head (@{$node->{headNode}}) {
	    foreach my $PT (@{$head->{possibleParents}}) {
		push(@{$node->{possibleParents}}, $PT) unless (FindInList($PT, $node->{possibleParents}));
	    }
	}
    } else {
	$noHead=1;
	warn("No head found for node $node->{sym}.\n");
    }

    if (!exists($node->{possibleParents}) || !defined($node->{possibleParents})) {
	if (exists($node->{PT})) {
	    $node->{possibleParents}=[$node->{PT}];
	} else {
	    $node->{possibleParents}=[];
	}
    } 

    $node->{possiblePTs}=[@{$node->{possibleParents}}];

    if (exists($node->{PT})) {
	unless (FindInList($node->{PT}, $node->{possiblePTs}) || $noHead) {
	    my @heads=map {$_->{sym}} sort {$a->{start} <=> $b->{start}} @{$node->{headNode}};
	    warn "PT $node->{PT} on node $node->{sym} inconsistent with the head-POS of the following heads: ", join(" ", @heads), ".\n";
	}
    }

    
    # Fourth, we figure out the what the PT ought to be as best we can.

    NarrowPT($node);

}# FindHeads

##############################

sub FindFrameHead {
    my ($frameNode)=@_;

    $frameNode->{headNode}=[];
    if ($#{$frameNode->{targetNodes}} > 0) {
	my @targetwordforms=map {$_->{form}} sort {$a->{start} <=> $b->{start}} @{$frameNode->{targetNodes}};
	warn "Possibly finding wrong head-target of multiword '", join(' ', @targetwordforms), "'.\n"; 
	foreach my $possibleHead (@{$frameNode->{targetNodes}}) {
	    $frameNode->{headNode}=TopRanked($frameNode->{headNode}, $possibleHead);
	}
    } else {
	$frameNode->{headNode}[0]=$frameNode->{targetNodes}[0];
    }
    if (exists($TransparentFE{$frameNode->{frameName}})) {
	foreach my $childedge (@{$frameNode->{childEdges}}) {
	    if (exists($childedge->{FE}) && $childedge->{FE} eq $TransparentFE{$frameNode->{frameName}}) {
		$childedge->{type}='SemHead';
	    }
	}
    }

}  # end FindFrameHead

sub FindGovHead {
    my ($govNode)=@_;
    my @governors = map {$_->{childNode}} grep {$_->{type} eq $govNode->{type}} @{$govNode->{childEdges}};
    
    if ($#governors > 0) {
	warn "Possibly finding wrong head Gov/Supp.\n";
	foreach my $gov (@governors) {
	    if ($PTtoType{$HeadPOStoPTs{$gov->{POS}}[0]} eq 'V') {
		$govNode->{headNode}=$gov;
		last;
	    } elsif (!exists($govNode->{headNode})) {
		$govNode->{headNode}=$gov;
	    }
	}
    } else {
	$govNode->{headNode}[0]=$governors[0];
    }
} # end FindGovHead

sub FindListHead {

    my($listNode)=@_;

    $listNode->{headNode}=[];

    if (!exists($listNode->{PT})) {
#	warn("No PT found on $listNode->{sym}:\n     ", NodeToText($listNode), "\nGuessing PT, probably wrongly\n");

	# Since we don't have anything to limit our search, we start with all of the childnodes as possible heads.
	# We then find the "Highest ranked" of these, according to V >> N, etc.
	
	foreach my $possibleHead (map {$_->{childNode}} @{$listNode->{childEdges}}) {
	    $listNode->{headNode}=TopRanked($listNode->{headNode}, $possibleHead);
	}
	
    } else {
	
	foreach my $possibleHead (map {$_->{childNode}} @{$listNode->{childEdges}}) {
#	    print "finding head-child of Node $listNode->{sym}...\n";

	    # Here we look through all the PTs mentioned as possible parents of the childnodes;
	    # the child nodes that have PTs that match the actual PT of the list are likely heads. 
	    
	    foreach my $pt (@{$possibleHead->{possibleParents}}) {
		if ($pt eq $listNode->{PT}) {
		    push (@{$listNode->{headNode}}, $possibleHead);
		    last;
		}
	    }
	}
    }

    # Mark the heuristically chosen heads with the edgetype 'Head'.

    foreach my $head (@{$listNode->{headNode}}) {
	$head->{parentEdge}{type}='Head' unless ($head->{parentEdge}{type} ne 'NonHead');
    }
} 

# This routine finds heads by process of elimination.

sub NarrowHeads {
    my($node)=@_;
    my @heads = [map {$_->{childNode}} @{$node->{childEdges}}];
    my $i=$#heads;
    while ($i>0) {
	if (exists($heads[$i]{POS}) && exists($NonHeadPOStoPTs{$heads[$i]{POS}})) {
	    splice(@heads, $i, 0);
	}
	$i--;
    }
    if ($#heads==0) {
	$node->{headNode}=[@heads];
    }
}


####################################################################

sub NodeToText {
    my($node, $feature)=@_;
    $feature='form' unless (defined($feature));
    
    my @words = grep {$_->{start} >= $node->{start} && $_->{end} <= $node->{end}} 
                     values %StartToWord;
    my $text= join(' ', map {$_->{$feature}} sort {$a->{start} <=> $b->{start}} @words);
    return $text;
}

#####################################################################
#
# This routine returns the highest-ranking wordnode(s), first by Supp/Cop vs. Target and then under the 
# following order: V, Prep, N, A (encoded in %TypeRanking); this allows us to predict that the
# syntactic head of "be on solid ground" is 'be'.

sub TopRanked {
    my($nodeList, $node2)=@_;
        
    if ((!exists($nodeList->[0]{parentEdge})) || 
	 ($nodeList->[0]{parentEdge}{type} eq 'Target' && 
	  ($node2->{parentEdge}{type} eq 'Supp' || $node2->{parentEdge}{type} eq 'Cop'))) {
	return [$node2];
    } elsif ($node2->{parentEdge}{type} eq 'Target' && 
	     ($nodeList->[0]{parentEdge}{type} eq 'Supp' || $nodeList->[0]{parentEdge}{type} eq 'Cop')) {
	return $nodeList;
    } else {
	my $PT1;
	my $PT2;

	if (!exists($node2->{POS}) || 
	    !exists($HeadPOStoPTs{$node2->{POS}}) ||
	    !exists($PTtoType{$HeadPOStoPTs{$node2->{POS}}[0]}) ||
	    !exists($TypeRanking{$PTtoType{$HeadPOStoPTs{$node2->{POS}}[0]}})) {
	    if (exists($node2->{PT}) && exists($TypeRanking{$PTtoType{$node2->{PT}}})) {
		$PT2=$node2->{PT};
	    } else {
		dmsg("Couldn't find rank of ", NodeToText($node2), " on ", $node2->{sym},"\n");
		return $nodeList;
	    }
	} else {
	    $PT2= $HeadPOStoPTs{$node2->{POS}}[0];
	}

	if (!exists($nodeList->[0]{POS}) || 
	    !exists($HeadPOStoPTs{$nodeList->[0]{POS}}) ||
	    !exists($PTtoType{$HeadPOStoPTs{$nodeList->[0]{POS}}[0]}) ||
	    !exists($TypeRanking{$PTtoType{$HeadPOStoPTs{$nodeList->[0]{POS}}[0]}})) {
	    if (exists($nodeList->[0]{PT}) && exists($TypeRanking{$PTtoType{$nodeList->[0]{PT}}})) {
		$PT1=$nodeList->[0]{PT};
	    } else {
		dmsg("Couldn't find rank of ", NodeToText($nodeList->[0]), " on ", $nodeList->[0]{sym},"\n");
		return [$node2];
	    }
	} else {
	    $PT1= $HeadPOStoPTs{$nodeList->[0]{POS}}[0];
	}

	if (exists($PTtoType{$PT1}) && exists($PTtoType{$PT2})) {
	    if ($TypeRanking{$PTtoType{$PT1}} > $TypeRanking{$PTtoType{$PT2}}) {
		return $nodeList;
	    } elsif ($TypeRanking{$PTtoType{$PT1}} < $TypeRanking{$PTtoType{$PT2}}) {
		return [$node2];
	    }
	} else {
	    dmsg("No type information for ", NodeToText($nodeList->[0]), " with $PT1 or ", NodeToText($node2) ," with $PT2");
	}
	return [@{$nodeList}, $node2];
    }
}


######################################################################
#
# This routine shouldn't be post-processing, but rather should be SOP
# whenever making a new word node.
#

sub AddWordPT {
    my($wordnode)=@_;
    
    if (exists($wordnode->{possibleParents})) {
	return;
    }
    if (exists($HeadPOStoPTs{$wordnode->{POS}})) {
	push @{$wordnode->{possibleParents}}, @{$HeadPOStoPTs{$wordnode->{POS}}};
    } elsif (exists($wordnode->{PT})) {
	push @{$wordnode->{possibleParents}}, $wordnode->{PT};
    }
    if (exists($FormsToGFs{$wordnode->{form}})) {
	if (exists($wordnode->{parentEdge}{GF}) && 
	    $wordnode->{parentEdge}{GF} ne $FormsToGFs{$wordnode->{form}}) {
	    warn("Contradictory GF above wordnode", $wordnode->{sym}, "\n");
	} else {
	    $wordnode->{parentEdge}{GF}=$FormsToGFs{$wordnode->{form}};
	}
    }
    $wordnode->{possiblePTs}=$wordnode->{possibleParents};

    # PTs are not assigned to words if they don't have them, but we do check to see if the 
    # word is labeled with a PT and if that is consistent with the POS.

    if (exists($wordnode->{PT}) && !FindInList($wordnode->{PT}, $wordnode->{possiblePTs})) {
	dmsg("POS $wordnode->{POS} inconsistent with PT $wordnode->{PT} on node $wordnode->{sym}, ' $wordnode->{form} '.\n");
    }

#   $Data::Dumper::Maxdepth=3;
#   print Dumper($wordnode);

}# End AddWordPT

#################################################################################################


sub NarrowPT {
    my ($node) = @_;

    GFFilterPTs($node);

    if ($#{$node->{possiblePTs}} > -1 && $node->{possiblePTs}[0] ne 'None') {
	my $ptHypothesis= $node->{possiblePTs}[0];
	if ($#{$node->{possiblePTs}} > 0 ) {
	    dmsg("Indeterminate PT on $node->{sym}\n");
	}

	if (!exists($node->{PT})) {
	    dmsg("guessing PT $ptHypothesis on Node $node->{sym}...\n");
	    $node->{PT}=$ptHypothesis;
	} elsif ($node->{PT} ne $ptHypothesis) {
	    dmsg("PT/GF/POS inconsistencies on or below Node $node->{sym}...\n");
	}

    } else {
	dmsg("$node->{sym} appears not to be a head\n");
    }
 
}# NarrowPT

sub GFFilterPTs {
    my ($node)=@_;

    my %GFFound=%{FindMajorGFs($node)};
    foreach my $gf (keys %GFtoCheck) {
	for (my $n=$#{$node->{possiblePTs}}; $n>-1; $n--) {
		dmsg("PT hypothesis $n for $node->{sym}: $node->{possiblePTs}[$n]\n");
		if (!$PTrequiresGF{$node->{possiblePTs}[$n]}{$gf} && exists($GFFound{$gf})) {
		    dmsg("$node->{possiblePTs}[$n] didn't match permissible PT for GF $gf; removing\n");
		    splice (@{$node->{possiblePTs}}, $n, 1);
		}
	    } 

	# This check limited to frame nodes, since even if
	# there happens to be GF information on non-frames, it 
	# is only on frames that all the appropriate GF info is 
	# obligatorily present, i.e. absence of a GF is informative.

	if ($node->{type} eq 'Frame' && !exists($GFFound{$gf})) {
	    dmsg("Didn't find GF $gf; deleting PTs that require $gf\n");
	    
	    for (my $n=$#{$node->{possiblePTs}}; $n>-1; $n--) {
		    
		# if GF required for PT, then remove PT; "S"s require Ext
		
		if (exists($PTrequiresGF{$node->{possiblePTs}[$n]}{$gf}) && 
		    $PTrequiresGF{$node->{possiblePTs}[$n]}{$gf}>.5) {
		    dmsg("$node->{possiblePTs}[$n] requires $gf; removing\n");
		    splice (@{$node->{possiblePTs}}, $n, 1);
		}
	    }
	}    
    }    
}


sub FindMajorGFs {

    my($node)=@_;
    
    # For any GFs which are looked for and found, this hash stores the local constituent with that GF; 
    # this enables filtering so that PTs which require the GF can be filtered out if GF not found. 
    
    my %GFFound=();
    
    foreach my $childConstit (@{$node->{childEdges}}) {
	if (!exists($childConstit->{type})) {
	    print "WTF!:\n", Dumper($childConstit);
	} elsif (exists($childConstit->{Other}) && $childConstit->{Other} eq 'Null') {

	    # Since a Null may count as Ext, this acts as if it's found Ext without splicing out possibilities

	    $GFFound{Ext}=$childConstit unless (exists($GFFound{Ext}));

	}
	if (exists($childConstit->{GF}) && $GFtoCheck{$childConstit->{GF}}) {
	    $GFFound{$childConstit->{GF}}=$childConstit;
	}
    }
    return \%GFFound;

} # end sub FindMajorGFs


######################################################################
#
# For every non-local edge, move the destination as high in the
# tree as possible (e.g. from target to frame if the spans are 
# the same).
#

sub PostProcessMoveUp {
    my($nledge, $oldnode, $newnode, $start, $end);
    foreach $nledge (FindEdges($SentenceNode, sub { exists($_[0]->{metatype}) && $_[0]->{metatype} eq 'nlEdge'; })) {
	$oldnode = $nledge->{childNode};
	$start = $oldnode->{start};
	$end = $oldnode->{end};
	$newnode = $oldnode;
	while ($newnode->{parentEdge}{parentNode}{start} == $start &&
	       $newnode->{parentEdge}{parentNode}{end} == $end) {
	    $newnode = $newnode->{parentEdge}{parentNode};
	}
	if ($newnode ne $oldnode) {
	    RemoveFromList($oldnode->{nlParentEdges}, $nledge);
	    $nledge->{childNode} = $newnode;
	    InsertEdge($newnode->{nlParentEdges}, $nledge);
	}
    }
} # PostProcessMoveUp()

######################################################################
#
# Any list which only has one child can be removed. Labels
# on the list (e.g. phrase type, part of speech) have to be
# reconciled with the child.
#

sub PostProcessRemoveSingletonLists {
    my(@nodes);
    my(%ignorekeys, $parent, $child, $node, $key);

    # These keys are not copied from parent to child.
    map { $ignorekeys{$_} = 1; } qw { start end type childEdges nlChildEdges parentEdge nlParentEdges sym };
    
    @nodes = FindNodes($SentenceNode,
		       sub { 
			   $_[0]->{type} eq 'List' &&  
			       $#{$_[0]->{childEdges}} == 0
			       && (!exists($_[0]->{PT}) || $_[0]->{PT} ne '<s>')
#			       && $#{$_[0]->{nlChildEdges}} == -1
			   });

    foreach $node (@nodes) {

	# First we make sure that we wouldn't be trying to put nlChildEdges on a Word:

	if (exists($node->{nlChildEdges}) && $#{$node->{nlChildEdges}} >=0 && 
	    $node->{childEdges}[0]{childNode}{type} eq 'Word') {
	    next;
	}

	# Then we make sure that we're not removing a parentless node, i.e. a Top;
	# if not, we get the parent of the node:

	if (exists($node->{parentEdge}) && defined($node->{parentEdge}) && ref($node->{parentEdge})) {
	    $parent = $node->{parentEdge}{parentNode};
	} else {
	    warn "Can't remove top of tree\n";
	    next;
	}

	$child = $node->{childEdges}[0]->{childNode};

	# Copy over any labels (e.g. PT, POS). For now, they must either
	# not exist in the child or be identical in the child
	
	foreach $key (keys %$node) {
	    next if exists($ignorekeys{$key});
	    if (!exists($child->{$key})) {
		$child->{$key} = $node->{$key};
	    } elsif ($child->{$key} ne $node->{$key}) {
		warn("PostProcessRemoveSingletonLists() found non-matching $key values on $node->{sym} with $node->{$key} vs. $child->{sym} with $child->{$key} on the following:\n     ", NodeToText($child), "\n");
		$child->{$key} = $node->{$key};
	    }
	}

	# Make any non-local parent edges of the node point to its child.

	foreach my $nlpe (@{$node->{nlParentEdges}}) {
	    $nlpe->{childNode} = $child;
	    InsertEdge($child->{nlParentEdges}, $nlpe);
	}
	
	# Make any non-local child-edges of $node point to the child as their new parent.

	foreach my $nlce (@{$node->{nlChildEdges}}) {
	    $nlce->{parentNode} = $child;
	    InsertEdge($child->{nlChildEdges}, $nlce);
	}

	# Link the child to the parent and delete the node.

	$child->{parentEdge} = $node->{parentEdge};
	$child->{parentEdge}{childNode} = $child;
	RemoveFromList(\@AllNodes, $node);
    }
} # PostProcessRemoveSingletonLists()

######################################################################
#
# For each node that has non-local parents, figure out where the node
# should go.
#

sub PostProcessMoveLocals {
    my($unmovableGFs)=@_;
    my($didwork);
    $didwork = 0;

    # the following equiv to:
    #  return ($#grep {$_!=0} map {PostProcessMoveLocalNode($_, $unmovableGFs)} FindNodes($SentenceNode, sub { $#{$_[0]->{nlParentEdges}} >= 0 }))+1;
    # ... but more comprehensible.

    foreach my $node (FindNodes($SentenceNode, sub { $#{$_[0]->{nlParentEdges}} >= 0 })) {
	if (PostProcessMoveLocalNode($node, $unmovableGFs)) {
	    $didwork = 1;
	}
    }
    return $didwork;
} # PostProcessMoveLocals()

###################################################################

sub PostProcessMoveGoverned {
    my($unmovableGFs)=@_;
    my($didwork);
    $didwork = 0;

    # the following equiv to:
    #  return ($#grep {$_!=0} map {PostProcessMoveLocalNode($_, $unmovableGFs)} FindNodes($SentenceNode, sub { exists($_[0]->{governingNodes}) && $#{$_[0]->{governingNodes}} >= 0 }))+1;
    # ... but more comprehensible.

    foreach my $node (FindNodes($SentenceNode, sub { exists($_[0]->{governingNodes}) && 
							 $#{$_[0]->{governingNodes}} >= 0 })) {
	if (PostProcessLocalizeGoverned($node, $unmovableGFs)) {
	    $didwork = 1;
	}
    }
    return $didwork;
} # PostProcessMoveGoverned()

sub PostProcessLocalizeGoverned {
    my($node, $unmovableGFs)=@_;
    my $tempNode=$node;

    while (exists($tempNode->{parentEdge}) && defined($tempNode->{parentEdge}) &&
	   exists($tempNode->{parentEdge}{parentNode})) {
	if (PostProcessMoveLocalNode($node, $unmovableGFs)) {
	    return 1;
	}
	$tempNode=$tempNode->{parentEdge}{parentNode};
    }
    return 0;

}# end PostProcessMoveGoverned


#################################################################################################

sub PostProcessMoveLocalNode {
    my($node, $unGFs) = @_;
    my($start, $end, $nlnode, $nledge, $parent);
    my %unmovableGFs = map{$_=>1} @{$unGFs}; 

    foreach $nledge (@{$node->{nlParentEdges}}) {

	# Heads and Appositives shouldn't be moved under.
	if (exists($nledge->{GF}) && 
	    exists($unmovableGFs{$nledge->{GF}}) || 
	    exists($nledge->{Other}) &&
	    $nledge->{Other} eq 'Ant') {
	    return 0;
	}
    }
#    print("Trying move-locals on $node->{sym}\n");
 
    my @FEParents=@{$node->{nlParentEdges}};
    @FEParents=map {$_->{parentNode}} @FEParents;
    

    my $tempnode=$node;
#    print "Finding LCA for node $node->{sym}\n";
    my $localTop=FindLCA(\@FEParents, $node);


    while ($tempnode->{sym} ne $localTop->{sym}) {
#	print ("Traversing node $tempnode->{sym}\n");
	if (exists($tempnode->{parentEdge}{FE}) || exists($tempnode->{parentEdge}{reversedEdge}) || $tempnode->{parentEdge}{type} eq 'SemHead') {
	    push(@FEParents, $node->{parentEdge}{parentNode});
#	    print("Found FE-edge, so position is not completely free.\n");
	    last;
	}
	$tempnode=$tempnode->{parentEdge}{parentNode};
    }


#    print "Finding highest direct parent\n";

    my @parents=FindHighest($node, \@FEParents, {});

#    warn "Highest parent(s): ", GiveSyms(@parents), "\n\n";
    return 0 if (!exists($parents[0]) || !defined($parents[0]) || ref($parents[0]) ne 'HASH');
    if ($#parents==0 && defined($parents[0])) {
	$parent=$parents[0];
    } else {    
	foreach $nlnode (@parents) {
	    if (!exists($nlnode->{sym})) {
		#$Data::Dumper::Terse = 1;
		#$Data::Dumper::Maxdepth = 3;
		#print Dumper($nledge), "\n";
		error("nlnode with no sym from nledge $nledge->{sym}");
		return 0;
	    }
	    if (!defined($start) || $nlnode->{start} < $start) {
		$start = $nlnode->{start};
		$parent = $nlnode;
	    }
	    if (!defined($end) || $nlnode->{end} > $end) {
		$end = $nlnode->{end};
	    }
	}
	while ($parent->{end} < $end && exists($parent->{parentEdge})) {
	    $parent = $parent->{parentEdge}{parentNode};
	}
    }

    my %LightWords=map {$_=>1} qw(DT, RB, MD);
    
    if ($node->{end}+2 == $parent->{start} ||
	$node->{start}-2 == $parent->{end}) {
	if (MoveLocal($node, $parent)) {
#	    warn("Moved $node->{sym} to $parent->{sym}\n");
	    if (exists($node->{governingNodes})) {
		$node->{governingNodes}=[];
	    }
	    return 1;
	}


    # We should also move local if only punctuation or auxiliaries intervene, but that's hard...

    } elsif (
	     exists($StartToWord{$node->{end}+2}) && 
	     exists($StartToWord{$node->{end}+2}{POS}) && 	     
	     exists($Punct{$StartToWord{$node->{end}+2}{POS}}) && 
	     $StartToWord{$node->{end}+2}{end}+2 == $parent->{start}) {
	MakeNLEdge($node, $StartToWord{$node->{end}+2});
	if (MoveLocal($StartToWord{$node->{end}+2}, $node)) {
#	    warn("Moved punctuation $StartToWord{$node->{end}+2}{sym} to $node->{sym} \n");
	    return 1;
	}

    } elsif (
	     exists($EndToWord{$node->{start}-2}) && 
	     exists($EndToWord{$node->{start}-2}{POS}) && 	     
	     exists($Punct{$EndToWord{$node->{start}-2}{POS}}) && 
	     $EndToWord{$node->{start}-2}{start}-2 == $parent->{end}) {
	MakeNLEdge($node, $EndToWord{$node->{start}-2});
	if (MoveLocal($EndToWord{$node->{start}-2}, $node)) {
#	    warn("Moved punctuation $EndToWord{$node->{start}-2}{sym} to $node->{sym}\n");
	    return 1;
	}
    }

    return 0;
} # PostProcessMoveLocalNode()


######################################################################
#
# Called by postprocessing step. Move $node so that it is a child
# of $newparent.
#

sub MoveLocal {
    my($node, $newparent) = @_;
    my($gotone, $nledge, $edge);

    $Data::Dumper::Maxdepth=4;

    $gotone = 0;
    foreach my $tmpnledge (@{$node->{nlParentEdges}}) {
	if ($tmpnledge->{parentNode} eq $newparent) {
	    $nledge = $tmpnledge;
	    $gotone = 1;
	    last;
	}
    }

    if (!$gotone) {

	# Note:  the following should happen only when called
	# to move punctuation.
	warn("Had to construct edge from $newparent->{sym} to $node->{sym}, but dunno why \n");
	$nledge=MakeNLEdge($newparent, $node);

    }

    if ($newparent->{type} eq 'Word') {

	# Words can't be parents of anything, so we have to do something
	# if we're handed a word as parentnode

	$newparent=MakeNodeAboveWord($newparent, $nledge);
	$nledge->{parentNode}=$newparent;	
	push(@{$newparent->{nlChildEdges}}, $nledge);

    }



    # Delete the old local edge

    $edge = $node->{parentEdge};
    if (!RemoveFromList($edge->{parentNode}{childEdges}, $edge)) {
	print ("Couldn't find parent edge in list of parent's kids!\n");
    }
    # Have to delete an empty parent if it is created by the move

    if ($#{$edge->{parentNode}{childEdges}}<0) {

	foreach my $oldNLchildEdge (@{$edge->{parentNode}{nlChildEdges}}) {
	    $oldNLchildEdge->{parentNode}=$node;
	}
	push(@{$node->{nlChildEdges}}, @{$edge->{parentNode}{nlChildEdges}});

	foreach my $oldNLparentEdge (@{$edge->{parentNode}{nlParentEdges}}) {
	    $oldNLparentEdge->{childNode}=$node;
	}
	push(@{$node->{nlParentEdges}}, @{$edge->{parentNode}{nlParentEdges}});

	RemoveFromList($edge->{parentNode}{parentEdge}{parentNode}{childEdges}, $edge->{parentNode}{parentEdge});
	UpdateStartEnd($edge->{parentNode}{parentEdge}{parentNode});
    } else {
	UpdateStartEnd($edge->{parentNode});
    }
    
    # Make non-local edge local.

#    warn "Post-move node:  ", Dumper($node);
#    warn "Parent:  ", Dumper($newparent);
#    warn "Edge:  ", Dumper($nledge);

    if (!RemoveFromList($node->{nlParentEdges}, $nledge)) {
	print ("Bad child\n");
    }

    if (!RemoveFromList($newparent->{nlChildEdges}, $nledge)) {
	print("Node $newparent->{sym} did not refer to non-local edge $nledge->{sym}.\n");
    }
    $nledge->{metatype} = 'edge';
    InsertChildEdge($newparent, $nledge);
    $node->{parentEdge} = $nledge;
    return 1;
} # MoveLocal()



sub MakeNodeAboveWord {
    my($newparent, $nledge)=@_;
    my $wordparent=$newparent;
    
    if ($wordparent->{parentEdge}{type} eq 'Target' ||
	$wordparent->{parentEdge}{type} eq 'Supp' ||
	$wordparent->{parentEdge}{type} eq 'Cop' ||
	$wordparent->{parentEdge}{type} eq 'Ctrlr' ||
	$wordparent->{parentEdge}{type} eq 'Gov' ||
	$wordparent->{parentEdge}{type} eq 'Head') {
	
	# If we ended up here, we need to displace the $nledge from the Word
	# as parent to the Word's parent.  
	
	$newparent=$wordparent->{parentEdge}{parentNode};
	
    } else {
	
	# If we are told to move something under a non-Head word, life is hard.
	# First, we make a new $newparent, a List with the same span, PT, and parent
	# as the old $newparent i.e. $wordparent.  Then we insert the Word below it by modifying 
	# the List's child-edges and the Word's parent-edges.  
	
	$newparent = CreateNewNode($wordparent->{start}, $wordparent->{end}, $wordparent->{parentEdge}{parentNode}); 
	$newparent->{type}='List';
	$newparent->{PT}=$wordparent->{PT} if (exists($wordparent->{PT})); 
	
	$newparent->{nlParentEdges}=$wordparent->{nlParentEdges};
	foreach my $nonlocal (@{$newparent->{nlParentEdges}}) {
	    $nonlocal->{childNode}=$newparent;
	}
	$wordparent->{nlParentEdges}=[];
	
	$newparent->{childEdges}[0]=$wordparent->{parentEdge}; 
	$wordparent->{parentEdge}{parentNode}=$newparent;
    }
    RemoveFromList($wordparent->{nlChildEdges}, $nledge);	
    
    return $newparent;
}

##################################
#
# This routine takes a list of nodes ($nodeList together with $startNode)
# that are all located in the same graph underneath a single highest node
# and finds which node of the entire graph is the lowest common ancestor of 
# the list
#


sub FindLCA {
    my($nodeList, $startNode)=@_;
    my(@ancestorNodes, $LCA);

    $LCA=$startNode;
    while (exists($startNode->{parentEdge}) && exists($startNode->{parentEdge}{parentNode})) {
#	warn ("Traversing node $startNode->{sym}\n");
	push(@ancestorNodes, $startNode->{parentEdge}{parentNode});
	$startNode=$startNode->{parentEdge}{parentNode};
    }

    my @nodeListCopy=(@{$nodeList});
    foreach my $node (@nodeListCopy) {
	while (!FindInList($node, \@ancestorNodes)) {
	    $node=$node->{parentEdge}{parentNode};
	}
	$LCA=$node if (FindPosition($node, \@ancestorNodes) > FindPosition($LCA, \@ancestorNodes));
    }
#    print "From ", join(', ', map {$_->{sym}} @ancestorNodes), ", lowest common node of parents ", join(', ', map {$_->{sym}} @{$nodeList}), " is $LCA->{sym}\n";
    return $LCA;
}

###########################################################
#
# This routine just finds the position of an item in a list
# and returns that position as a number.  If it is not 
# present, returns -1.


sub FindPosition {
    my($item, $list)=@_;
    my $index=$#{$list};
    while ($index>=0) {
	last if ($item eq $list->[$index]);
	$index--;
    }
    return $index;
}


###################################

sub GiveSyms {
    my @nodes=@_;
    my $syms;
    if ($#nodes>=0 && defined($nodes[0]) && ref($nodes[0]) eq 'HASH') {
	$syms=join(', ', map {$_->{sym}} @nodes);
	return $syms;
    } else {
	return "undefined";
    }
}

######################################################################
#
# The following routine starts with a $node, and traverses the graph
# upward (via nonlocal and appropriate local edges), checking to see 
# which nodes are a member of the $list.  If there are no matches, 
# returns undef.  If there are several highest nodes, returns 0.  
# Otherwise, returns highest accessible $node included in $list.
#
# Each iteration up the graph checks to see if there are parents.  
# If so, iterates across the parents appending their values for 
# &FindHighest.  If any value is found among these parents, don't
# bother checking the current $node to see if it's in the list,
# since there are higher list members present.  If there are no higher,
# return self if $node is on $list, otherwise return undef.


sub FindHighest {
    my ($node, $list, $visited)=@_;
    my @matches=();
    my @parents=();

    if (exists($visited->{$node->{sym}})) {
	if ($visited->{$node->{sym}}==1) {
#	    warn "Hit loop in graph at $node->{sym}; ignoring...\n";
	    $visited->{$node->{sym}}=[];
	} else {
#	    warn "Already visited node $node->{sym}, returning ", GiveSyms(@{$visited->{$node->{sym}}}), "\n";
	}
	@matches=@{$visited->{$node->{sym}}};
#	return @{$visited->{$node->{sym}}};

    } else {

	$visited->{$node->{sym}}=1;
#	warn "Nodes visited so far: ", join(', ', keys %{$visited}), "\n";
	
	if (exists($node->{parentEdge}) && 
	    exists($node->{parentEdge}{parentNode})) {
	    push(@parents, $node->{parentEdge}{parentNode});
	}
	
	if (exists($node->{nlParentEdges}) && $#{$node->{nlParentEdges}} >=0) {
	    push(@parents, map {$_->{parentNode}} @{$node->{nlParentEdges}});
	}
	
	foreach my $parent (@parents) {
#	    my @temp=FindHighest($parent, $list, $visited);
#	    for (my $member=$#temp;$member>=0;$member--) {
#		if (FindInList($temp[$member], \@matches)) {
#		    splice(@temp, $member, 1);
#		}
#		
#		if (!keys %{$temp[$member]}) {
#		    error("Somehow got an empty hash from traveling above $node->{sym}\n");
#		    splice(@temp, $member, 1);
#		}
#	    }
#	    push(@matches, @temp) if ($#temp>=0 && ref($temp[0]) eq 'HASH');


	    foreach my $member (FindHighest($parent, $list, $visited)) {
		if (!FindInList($member, \@matches) &&
		    ref($member) eq 'HASH') {
		    push(@matches, ($member)) ;
		}
	    }
	}
	
	if (($#matches<0 || ref($matches[0]) ne 'HASH') && grep {$_ eq $node} @{$list}) {
	    $matches[0]=$node;
	}

	if ($#matches<0) {
#	    warn "No matches at or above node $node->{sym}\n";
	    $visited->{$node->{sym}}=[];
	    @matches=@{$visited->{$node->{sym}}};
#	    return @{$visited->{$node->{sym}}};
	} else {
#	    warn "Returning: ", GiveSyms(@matches), " from test of $node->{sym} and higher\n";
	    $visited->{$node->{sym}}=\@matches;
#	    return @matches;
	}
    }
    return @matches;
}


######################################################################

sub PostProcessRemoveSingletonGovs {
    my($gov, $nledge, $pedge, $cedge);
    foreach $gov (FindNodes($SentenceNode, sub { $_[0]->{type} eq 'Gov' && $#{$_[0]->{childEdges}} == 0 && $_[0]->{childEdges}[0]{childNode}{type} eq 'Frame'; })) {
	# Remove non-local edges
	foreach $nledge (@{$gov->{nlChildEdges}}) {
	    RemoveFromList($nledge->{childNode}{nlParentEdges}, $nledge);
	    delete($nledge->{childNode});
	    delete($nledge->{parentNode});
	}
	$pedge = $gov->{parentEdge};
	$cedge = $gov->{childEdges}[0];
	$pedge->{childNode} = $cedge->{childNode};
	$cedge->{childNode}{parentEdge} = $pedge;
	RemoveFromList(\@AllNodes, $gov);
    }
}

######################################################################
#
# Given a code ref, find all nodes for which the code ref
# returns true. The code ref is passed the node and all
# of the @rest arguments.
#
# This isn't very efficient (since it isn't tail-recursive).
#

sub FindNodes {
    my($top, $coderef, @rest) = @_;
    my(@nodes, $edge);

    # Initialize the graph-recorder, if necessary

    if (ref($rest[$#rest]) ne 'HASH' || !exists($rest[$#rest]{snog})) {
	push(@rest, {snog=>[]});
    }

    # If we come across a node twice, it can't be good.

    if (FindInList($top, $rest[$#rest]{snog})) {
	error("Loop in graph involving ", GiveSyms(@{$rest[$#rest]{snog}}), "\n");
	return undef;
    }

    # Push every node we visit into the graph-recorder

    push(@{$rest[$#rest]{snog}}, $top);
#    warn("So far searched: ", GiveSyms(@{$rest[$#rest]{snog}}), "\n");

    if (&$coderef($top, @rest)) {
	push(@nodes, $top);
    }
    foreach $edge (@{$top->{childEdges}}) {
	push(@nodes, FindNodes($edge->{childNode}, $coderef, @rest));
    }
    return @nodes;
} # FindNodes()


######################################################################
#
# Given a code ref, find all edges (both local and non-local) for
# which the code ref returns true. The code ref is passed the edge and
# all of the @rest arguments.
#
# This isn't very efficient (since it isn't tail-recursive).
#

sub FindEdges {
    my($top, $coderef, @rest) = @_;
    my(@edges, $edge);
    foreach $edge (@{$top->{childEdges}}) {
	if (&$coderef($edge, @rest)) {
	    push(@edges, $edge);
	}
	push(@edges, FindEdges($edge->{childNode}, $coderef, @rest));
    }
    foreach $edge (@{$top->{nlChildEdges}}) {
	if (&$coderef($edge, @rest)) {
	    push(@edges, $edge);
	}
    }
    return @edges;
} # FindEdges()

######################################################################
#
# Add POS to each word in a DG.
#

sub AddPOS {
    my($top) = @_;
    my(@wordnodes, $wordnode, $ifh, $ofh, $line, $pid, $i);
    $pid = open2($ifh, $ofh, $POS_TAGGER);
    @wordnodes = values %StartToWord;

    print $ofh join("\n", map { $_->{form};} @wordnodes), "\n";
    $ofh->close();
    $i = 0;
    while ($line = <$ifh>) {
	chop($line);
	$wordnodes[$i]->{POS} = $line;
	$i++;
    }
    waitpid($pid, 0);
} # AddPOS()

######################################################################
#
# 
#

sub ProcessAux {
    my ($node)=@_;
    my $found=0;
    my $limit;
    my $didwork=0;

    if ($node->{type} ne 'List' || exists($node->{PT}) && 
	exists($PTtoType{$node->{PT}}) && $PTtoType{$node->{PT}} ne 'V') {
	return 0;
    }

    foreach my $child (map {$_->{childNode}} @{$node->{childEdges}}) {
	if (exists($child->{form}) && exists($Aux{lc($child->{form})})) {
	    $child->{parentEdge}{type}='SynHead';
	    $found=1;
	    $limit=3;
	} elsif ($found) {
	    if ((exists($child->{POS}) && exists($HeadPOStoPTs{$child->{POS}}) && exists($PTtoType{$HeadPOStoPTs{$child->{POS}}[0]}) 
		 && $PTtoType{$HeadPOStoPTs{$child->{POS}}[0]} eq 'ADV') ||
		(exists($child->{PT}) && exists($PTtoType{$child->{PT}}) && $PTtoType{$child->{PT}} eq 'ADV')) {
		$limit--;
		if (!$limit) {
		    $found=0;
		}
	    } else {
		$found=0;
		if ((exists($child->{POS}) && exists($HeadPOStoPTs{$child->{POS}}) && exists($PTtoType{$HeadPOStoPTs{$child->{POS}}[0]}) &&
		     $PTtoType{$HeadPOStoPTs{$child->{POS}}[0]} eq 'V') ||
		    (exists($child->{PT}) && exists($PTtoType{$child->{PT}}) &&
		     $PTtoType{$child->{PT}} eq 'V')) {
		    $child->{parentEdge}{type}='SemHead';
		    $didwork=1;
		}
	    }
	}
    }
    
    return $didwork;
}

# Note that &ProcessConjunction checks to see if its conjuncts are marker nodes;
# if they are marker nodes, then &AddMarker is run on each, and IsMarkerNode will test false 
# on the parent node of the conjunction.

sub ProcessConjunction {
    my ($node)=@_;

    if($node->{type} ne 'List' || !exists($node->{PT})) {
	return 0;
    }

    my $didwork=0;
    my $i=1;
    my %match;
    my @matchBasis=('POS', 'PT', 'possibleParent');
    my $newNode;

    my @children=map {$_->{childNode}} @{$node->{childEdges}};
    while ($i<$#{$node->{childEdges}}) {
	if (exists($children[$i]{form}) && exists($Conjunction{lc($children[$i]{form})})) {
	    foreach my $key (@matchBasis) {
		if (exists($children[$i+1]{$key})) {
		    $match{$key}=$children[$i+1]{$key};
		}
	    }
	    my @chunkList=($i+1, $i);
	    for (my $j=$i-1; $j>=0; $j--) {
		if ((exists($children[$j]{POS}) && $children[$j]{POS} eq $match{POS}) ||
		    (exists($children[$j]{PT}) && $children[$j]{PT} eq $match{PT}) ||
		    (exists($children[$j]{possibleParent}) && $children[$j]{possibleParent} eq $match{possibleParent})) {
		    push(@chunkList, $j);
		} elsif (exists($children[$j]{form}) && exists($Punct{$children[$j]{form}})) {
		    next;
		} else {
		    if ($#chunkList > 1) {
			$newNode=RepartitionConjunction($node, [$i], \%match, \@chunkList);
			$i=$j+1;
			@children=map {$_->{childNode}} @{$node->{childEdges}};
			$didwork=1;
		    }
		    last;
		}
	    }
	}
	$i++;
    }
    return 0 if (!$didwork);
} # end ProcessConjunction

sub RepartitionConjunction {
    my ($oldParent, $conjunctions, $newInfo, $newChildren)=@_;
    my @childNodes=map { $oldParent->{childEdges}[$_]{childNode} } @{$newChildren};

    my $conjunctionNode=GetOrCreatNode($oldParent, $childNodes[$#childNodes]{start}, $childNodes[0]{end});

    my @newNodes=(GetOrCreateNode($oldParent, $childNodes[0]{start}, $childNodes[0]{end}));
    for (my $k=1; $k<=$#childNodes; $k++) {
	$newNodes[$k]=GetOrCreateNode($oldParent, $childNodes[$k]{start}, $childNodes[$k-1]{start}-2);
    }
    return $conjunctionNode;
}


sub AddMarker {
    my($pnode) = @_;
    my($ce, $cn, $pos, $i, $node, $limit);

    $limit=3;
    $limit=1 if ($pnode->{PT} eq 'Sfin');
    $limit=1 if ($pnode->{PT} eq 'Poss');

    # Poss phrases have a marker on the right side, so we do a right-to-left search.

    if ($pnode->{PT} eq 'Poss') {
	
	# Certain pronouns can make a Poss phrase all by themselves.
	
	if (exists($pnode->{POS}) && 
	    ($pnode->{POS} eq "PP\$" || $pnode->{POS} eq "WP\$")) {
	    return 0;
	}
	
	# Otherwise, the rightmost item should be a word with part-of-speech POS.
	
	for ($i = $#{$pnode->{childEdges}}; $i > -1 && $i >= $#{$pnode->{childEdges}} - $limit ; $i--) {
	    $ce = $pnode->{childEdges}[$i];
	    $cn = $ce->{childNode};
	    next if (!exists($cn->{POS}) || !defined($cn->{POS}));
	    $pos = $cn->{POS};
	    if ($pos eq 'POS' || (exists($cn->{form}) && $cn->{form} eq "'s")) {
		if ($ce->{type} ne 'NonHead') {
#		    error("Found an edge type that wasn't NonHead for marker $pnode->{sym} -> $cn->{sym}");
		    next;
		}
		if ($i <= 0) {
		    error("Found marker too far to the left; indicates wrong PT or boundary for:\n  ", NodeToText($pnode));
		    next;
		}
		$node = GetOrCreateNode($pnode->{start},
					$pnode->{childEdges}[$i-1]{childNode}{end});
		if (exists($node->{PT}) && $node->{PT} ne 'NP') {
		    error("Found a non NP phrase for marker under $pnode->{sym}\n");
		}
		$node->{PT} = 'NP';
		
		$pnode->{headNode}=[$cn];
		$pnode->{childEdges}[0]{type}='SemHead';
		$pnode->{childEdges}[0]{PT}=$node->{PT};

		if (IsMarkerNode($pnode)) {
		    $ce->{type} = 'Marker';
		    $pnode->{type}='MarkerPhrase';
		} else {
		    $ce->{type}='Head';
		}
		return 1;
	    }
	}

    } else {

	# All other marker-phrases have markers on the left, so we do a left-to-right search.
	
	dmsg("Looking for marker on node $pnode->{sym}.\n");

	for ($i = 0; $i <= $#{$pnode->{childEdges}} && $i <= $limit; $i++) {
	    $ce = $pnode->{childEdges}[$i];
	    $cn = $ce->{childNode};
	    next if (!exists($cn->{POS}));
	    $pos = $cn->{POS};
	    if ($pos eq 'IN' ||
		$pos eq 'RB' ||
		$pos eq 'TO' 
#		||
#		$pos eq 'PRF' ||
#		$pos eq 'PRP' ||
#		$pos eq 'AVP-PRP' ||
#		$pos eq 'CJS-PRP' ||
#		$pos eq 'TO0' 
		) {
		if ($ce->{type} ne 'NonHead') {
		    dmsg("Found an edge type that wasn't NonHead for marker $pnode->{sym} -> $cn->{sym}");
		    next;
		}
		if ($i < $#{$pnode->{childEdges}}) {
		    $node = GetOrCreateNode($pnode->{childEdges}[$i+1]{childNode}{start},
					    $pnode->{end});
		    
		    if (exists($MarkerToBare{$pnode->{PT}})) {
			if (exists($node->{PT}) && $node->{PT} ne $MarkerToBare{$pnode->{PT}}) {
			    warn("Found a non $MarkerToBare{$pnode->{PT}} under $pnode->{sym}\n");
			}
			$node->{PT}=$MarkerToBare{$pnode->{PT}};
		    } else {
			$node->{PT}='<unknown>' unless (exists($node->{PT}));
		    }
		    
		    $pnode->{headNode}=[$cn];
		    $pnode->{childEdges}[$#{$pnode->{childEdges}}]{type}='SemHead';
		    $pnode->{childEdges}[$#{$pnode->{childEdges}}]{PT}=$node->{PT};

		    if (IsMarkerNode($pnode)) {
			$ce->{type} = 'Marker';
			$pnode->{type} = 'MarkerPhrase';
			dmsg("Made $pnode->{sym} a MarkerPhrase.\n");
		    } else {
			$ce->{type}='Head';
		    }
		    return 1;
		}
	    }
	}
    }
    
    dmsg("Node $pnode->{sym} appears to be a marker parent, but no marker found.");
    return 0;

} # AddMarker()

sub IsMarkerNode {
    my($node) = @_;
    
    
    if (exists($node->{type}) && $node->{type} eq 'MarkerPhrase') {
	dmsg("$node->{sym} was already labeled a marker phrase.\n");
	return 0;
    }
    dmsg("Checking node $node->{sym} for markerhood.\n");

    foreach my $edge (@{$node->{nlParentEdges}}, $node->{parentEdge}) {
	if (exists($edge->{FE}) && exists($RelEdges{$edge->{FE}}) && 
	    (!exists($edge->{parentNode}{frameName}) || 
	     !exists($RelExceptions{$edge->{FE}}) || 
	     !exists($RelExceptions{$edge->{FE}}{$edge->{parentNode}{frameName}}))) {
	    dmsg("Non-marker semantics.\n");
	    return 0;
	}
    }
    dmsg("Marker found.\n");
    return 1;
} # IsMarkerNode()				   

######################################################################
#
# Because perl uses reference counting for garbage collection,
# circular (dead) data structures are not collected. To prevent
# memory leaks, users of dg.pm should call FreeDG on the top
# node of the graph when they're done with it.
#
# Just clears the edge pointers, which should break any cycles.

sub FreeDG {
    my($top) = @_;
    my($edge);
    foreach $edge (FindEdges($top, sub { 1; })) {
	delete($edge->{parentNode});
	delete($edge->{childNode});
    }
}

sub warning {
    print STDERR "Warning: ", join(' ', @_), "\n";
}

sub error {
    print "Error: ", join(' ', @_), "\n";
#    print(&stacktrace());
    $ErrorFlag = 1;
} # error();


sub stacktrace {
    my($ret, @caller, $i);
    $ret = '';
    for ($i = 1; 1; $i++) {
        @caller = caller($i);
        last if $#caller < 3;
        $ret .= "file: '$caller[1]' subroutine: '$caller[3]' ";
        @caller = caller($i-1);
        $ret .= "line: $caller[2]\n";
    }
    return $ret;
}

sub dmsg {
    if ($Debug) {
	print join(" ", @_), "\n";
    }
}


# The following is used for debugging. It generates a snapshot
# of the dg each time with the names snap01.dot, snap02.dot, etc.
# It also called dot to generate the pngs from the dots. This
# can be slow. 
#
# No checks are made for if the files exist already.

{
my $SnapCount = 0;
sub snapshot {
    my($node, $name) = @_;
    my($fname, $fh);
    if ($Debug > 0) {
	if (!defined($SnapCount)) {
	    $SnapCount = 0;
	}
	if (!defined($name)) {
	    $name = 'snap';
	}			     
	$fname = sprintf("$name%02d", $SnapCount);
	print STDERR "Snapshot $fname\n";
	$fh = new FileHandle(">$fname.dot") or die "Couldn't write $fname.dot: $!";
	print $fh &dgToDot($node);
	$fh->close();
	system("dot -Tpng $fname.dot > $fname.png");
	$SnapCount++;
    }
}
}


######################################################################
#
# Routines below here implement generation of a .dot file given
# a dg. Notice the new scope...
#
# $opts should be a hash ref. Valid opts currently are just
# 'print', which should contain a comma seperated list
# of things to print. If 'print' isn't provided, print everything.
# Valid 'print's are Dep start end ID sym PT GF POS FE GF Sent Other Constructions
#

{

my(@Nodes, @Edges);	# Array of strings; dot commands for drawing nodes and edges
my(%NLSymToN);          # For nodes that have non-local parents, store the number.
my($SymN, $DotSymIndex);
my(%PrintOpt);

sub dgToDot {
    my($top, $opts) = @_;
    my(@allprint);

    @allprint = qw (Dep start end ID sym PT GF POS FE GF Other Sent Cotarget);
    
    if (!defined($opts) || !exists($opts->{print}) || $opts->{print} eq 'all') {
	map { $PrintOpt{$_} = 1; } @allprint;
    } else {
	map { $PrintOpt{$_} = 0; } @allprint;
	map { $PrintOpt{$_} = 1; } split(/\s*[, ]+\s*/, $opts->{print});
    }

    $SymN = 0;
    $DotSymIndex = 0;
    %NLSymToN = ();
    @Nodes = ();
    @Edges = ();
    
    dgToDotNode($top);

    # There's a bit of ugliness here involving generation of the indirect nodes.
    # We don't know their label num until we've traversed the tree, so I just used
    # the node sym with a marker (REPLACE # n45 # REPLACE). Once we're done
    # traversing, we can just do a string replace.

    my $nodestr = join("\n", @Nodes);
    $nodestr =~ s/REPLACE \# (n[0-9]+) \# REPLACE/$NLSymToN{$1}/g;

    return "digraph {\nordering=out;\n\n$nodestr\n\n" . join("\n", @Edges) . "\n}\n";
}

sub dgToDotNode {
    my($node) = @_;
    my($childedge, $str, $sym, $labels);
    $str = '';

    # Output yourself and then your children.
    $str = "$node->{sym} [";

    # Check if we have non-local parents
    if ($#{$node->{nlParentEdges}} >= 0) {
	$SymN++;
	$NLSymToN{$node->{sym}} = $SymN;
	$str .= " shape=record, label=\"$SymN|";
    } elsif ($node->{type} eq 'Word') {
	$str .= " shape=box, label=\"";
    } else {
	$str .= " label=\"";
    }
    if (exists($node->{ID}) && $PrintOpt{ID}) {
	$str .= "ID: $node->{ID}\\n";
    }
    if ($PrintOpt{sym}) {
	$str .= "$node->{sym}\\n";
    }
    if ($PrintOpt{start} && $PrintOpt{end}) {
	$str .= "$node->{start} $node->{end}\\n";
    } elsif ($PrintOpt{start}) {
	$str .= "$node->{start}\\n";
    } elsif ($PrintOpt{end}) {
	$str .= "$node->{end}\\n";
    }
    $str .= "$node->{type}\\n";
    if ($Debug && exists($node->{possiblePTs})) {
	$str .= "posPTs:";
	foreach my $pt (@{$node->{possiblePTs}}) { 
	    $str .= " $pt";
	}
    }
    if (exists($node->{form})) {
	$str .= "\\n\'$node->{form}\'";
    }
    if (exists($node->{frameName})) {
	$str .= "\\n$node->{frameName}";
    }
    if (exists($node->{PT}) && $PrintOpt{PT}) {
	$str .= "\\nPT: $node->{PT}";
    }
    if (exists($node->{POS}) && $PrintOpt{POS}) {
	$str .= "\\nPOS: $node->{POS}";
    }
    if (exists($node->{denoted})) {
	$str .= "\\nDenoted: " . join(", ", @{$node->{denoted}});
    }
    $str .= "\" ];";

    push(@Nodes, $str);

    # First output nonlocal edges, then local edges. Note
    # that non-local edges do not cause recursion.
    
    foreach $childedge (@{$node->{nlChildEdges}}) {
	$sym = $DotSymIndex++;
	push(@Nodes, "nl$sym [ shape=box, color=lightgray, style=filled, width=.25, height=.25, label=\"REPLACE # $childedge->{childNode}{sym} # REPLACE\"];" );
        $str = "$node->{sym} -> nl$sym [ style=dotted";
	$labels = edgelabels($childedge);
	if (defined($labels)) {
	    $str .= ", label=\"$labels\"";
	}
	$str .= " ];";
	push(@Edges, $str);
    }

    # Now add local children
    
    foreach $childedge (@{$node->{childEdges}}) {
	dgToDotEdge($childedge);
    }
}
	
sub dgToDotEdge {
    my($edge) = @_;
    my($childnode, $str, $labels);

    # Output your child and yourself
    dgToDotNode($edge->{childNode});
    if (exists($edge->{reversedEdge})) {
	$str .= $edge->{parentNode}{sym} . " -> " . $edge->{childNode}{sym} . " [ dir=both, ";
    } else {
	$str .= $edge->{parentNode}{sym} . " -> " . $edge->{childNode}{sym} . " [ dir=forward, ";
    }
    $labels = edgelabels($edge);
    if (defined($labels)) {
	$str .= "label=\"$labels\" ]";
    }
    $str .= ";";
    push(@Edges, $str);
}

sub edgelabels {
    my($edge) = @_;
    my(@labels, $lab);
    @labels = ();
#    push(@labels, $edge->{sym});
    if (exists($edge->{type}) && ($edge->{type} ne 'Dep' || $PrintOpt{Dep})) {
	push(@labels, "$edge->{type}");
    } 
    foreach $lab (qw (FE GF PT Other Sent)) {
	if (exists($edge->{$lab}) && $PrintOpt{$lab}) {
	    push(@labels, "$lab: $edge->{$lab}");
	}
    }

    if ($PrintOpt{Cotarget}) {
	foreach $lab (qw ( Noun Adj Adv Prep Verb)) {
	    if (exists($edge->{$lab})) {
		push(@labels, "$lab: $edge->{$lab}");
	    }
	}
    }
    if (exists($edge->{reversedEdge})) {
	push(@labels, 'Upward:');
	foreach $lab (qw (FE GF PT Other Sent)) {
	    if (exists($edge->{reversedEdge}{$lab}) && $PrintOpt{$lab}) {
		push(@labels, "$lab: $edge->{reversedEdge}{$lab}");
	    }
	}
	if ($PrintOpt{Cotarget}) {
	    foreach $lab (qw ( Noun Adj Adv Prep Verb)) {
		if (exists($edge->{reversedEdge}{$lab})) {
		    push(@labels, "$lab: $edge->{$lab}");
		}
	    }
	}
    }
	
    if ($#labels >= 0) {
	return join("\\l", @labels) . "\\l";
    } else {
	return undef;
    }
}
}


1;
