#!/usr/bin/env perl

# ########################################################################
#
# fssScore.pl - Frame Semantic Structure Extraction Scorer
#
# ########################################################################

use strict;
use File::Basename;
use Getopt::Std;
use XML::Parser;
use vars qw($opt_d $opt_h $opt_v $opt_s);

# global variables
my $PROG = basename($0);
my $DEBUG = 0;
my $VERBOSE = 0;
my $OUTPUT_PER_SENTENCE = 0;
my $FRAME_POINTS = 1;
my $COREFE_POINTS = 1;
my $NONCOREFE_POINTS = .5;
my $SUPP_POINTS = .5;
my $NE_POINTS = .5;
my $SEPARATOR = ';';
my %CORELOOKUP;
my $PREC = 5;       # number of decimal digits in Precision, Recall, Fscore

# global variables used for printing matches
my $MISSED = 0;
my $MATCHED = 1;
my $EXTRA = 2;
my $FEATURE_WIDTH = 80;
my $SCORE_WIDTH = 15;

MAIN:{
    
    # parse options
    getopts("dhvs") || &usage();
    if (defined $opt_d) {
	$DEBUG = 1;
    }
    if (defined $opt_s) {
        $OUTPUT_PER_SENTENCE = 1;
    }
    if (defined $opt_v) {
        $VERBOSE = 1;
    }

    # read arguments
    if ($#ARGV!=2 || defined($opt_h)) {
	&usage();
    }
    my $framesFile = shift(@ARGV);
    my $f1 = shift(@ARGV);
    my $f2 = shift(@ARGV);

    my $f;
    foreach $f ($framesFile, $f1, $f2) {
        if (! -f $f) {
            die ("ERROR: could not read $f\n");
        }
    }
    
    # populate global CORELOOKUP (core vs. non-core)
    %CORELOOKUP = &CreateFELookup($framesFile);
    
    # read gold standard XML
    my ($goldWContent, $goldStand, $goldSentWL) = &parseXMLFile($f1);
    print "Done parsing $f1\n" if $DEBUG;

    # read the xml to score
    my ($scoreWContent, $toScore, $scoreSentWL) = &parseXMLFile($f2);
    print "Done parsing $f2\n" if $DEBUG;
 
    # do comparison / calculate scores

    my $sent;
    my $totalMatchSum = 0;
    my $totalScoreSum = 0;
    my $totalGoldSum = 0;
    my $nSent = 0;

    foreach $sent (sort keys %{$goldStand}) {
	my ($goldMS, $scoreMS);
        my ($goldWL, $scoreWL);
	my $goldSum = 0;
        my $scoreSum = 0;

	# check if sentences are missing, 0 points
	if (!defined($scoreWContent->{$sent})) {
	    print STDERR "WARNING: sentence $sent is missing in $f2.  Skipping\n";
	    next;
	}

	# calculate gold standard point total
	$goldMS = $goldStand->{$sent};
	$goldSum = &arraySum(values %{$goldMS});
	if ($goldSum==0) {
	    print STDERR "No gold standard annotation on sentence $sent. Skipping.\n";
	    next;
	}

	# comparison is valid only if the <words> section is identical
	if ($goldWContent->{$sent} ne $scoreWContent->{$sent}) {
	    print STDERR "WARNING: sentences for $sent differ in <words> content. Skipping.\n";
	    next;
	}

	print "sentences with ID $sent have identical <words>\n" if $DEBUG;
	$scoreMS = $toScore->{$sent};
	$scoreSum = &arraySum(values %{$scoreMS});

	# Calculate Precision and Recall and Fscore

	# Sum up points for gold standard features that are matched
	my $feature;
	my $matchSum = 0;

        # Column Labels for VERBOSE mode
        print(("-" x $FEATURE_WIDTH)."\n") if $VERBOSE;
        print "ID=$sent: " if $VERBOSE;
        &printSentence($goldSentWL->{$sent}) if $VERBOSE;
        &printMatch("ColumnLabels","",0) if $VERBOSE;

        # check for feature matches
        foreach $feature (keys %{$goldMS}) {
            if (defined $scoreMS->{$feature}) {
                print "matching feature: $feature ($scoreMS->{$feature} pts)\n" if $DEBUG;
                &printMatch(&prettyFeature($feature,$scoreSentWL->{$sent}),$scoreMS->{$feature},$MATCHED) if $VERBOSE;
                $matchSum += $scoreMS->{$feature};
            } else {
                print "missing feature: $feature ($goldMS->{$feature} pts)\n" if $DEBUG;
                &printMatch(&prettyFeature($feature,$goldSentWL->{$sent}),$goldMS->{$feature},$MISSED) if $VERBOSE;
            }
        }
        if ($DEBUG || $VERBOSE) {
            foreach $feature (keys %{$scoreMS}) {
                if (!defined($goldMS->{$feature})) {
                    &printMatch(&prettyFeature($feature,$scoreSentWL->{$sent}),$scoreMS->{$feature},$EXTRA) if $VERBOSE;
                    print "extra feature: $feature ($scoreMS->{$feature} pts)\n" if $DEBUG;
                }
            }
        }
        &printScoreTotal($goldSum,$matchSum,$scoreSum) if $VERBOSE;

        $totalMatchSum += $matchSum;
        $totalScoreSum += $scoreSum;
        $totalGoldSum += $goldSum;

        my $precision = $matchSum / $scoreSum;
	my $recall = $matchSum / $goldSum;
	my $fscore = (2 * $recall * $precision) / ($recall + $precision);

	if ($OUTPUT_PER_SENTENCE) {
            printf("Sentence ID=%d: Recall=%.${PREC}f (%.1f/%.1f) Precision=%.${PREC}f (%.1f/%.1f) Fscore=%.${PREC}f\n",
                   $sent, $recall, $matchSum, $goldSum, $precision, $matchSum, $scoreSum, $fscore);
        }
	$nSent++;
    }
    my $totalRecall = $totalMatchSum / $totalGoldSum;
    my $totalPrecision = $totalMatchSum / $totalScoreSum;
    my $fScore = (2 * $totalPrecision * $totalRecall ) / ($totalPrecision + $totalRecall);
    
    printf("%d Sentences Scored: Recall=%.${PREC}f (%.1f/%.1f)  Precision=%.${PREC}f (%.1f/%.1f)  Fscore=%.${PREC}f\n",
           $nSent,$totalRecall, $totalMatchSum, $totalGoldSum,
           $totalPrecision, $totalMatchSum, $totalScoreSum, $fScore);

    exit;
}
# end MAIN

# ################################################################
# ################################################################
# ################################################################
# ################################################################

#
# Returns sum of numbers in an array
#
sub arraySum {

    my $sum = 0;
    my $num;
    foreach $num (@_) {
	$sum += $num;
    }
    return $sum;
}

#
# prints a sentence's words given a sentence hash (indexed by node id)
#
sub printSentence {

    my ($wHash) = @_;

    my $node;
    my $sent;
    foreach $node (sort keys %{$wHash}) {
        if ($sent) {
            $sent .= " ".$wHash->{$node};
        } else {
            $sent = $wHash->{$node};
        }
    }
    print $sent . "\n";
}

#
# Replace node labels with their corresponding words to make a 
# pretty feature string for printing.
#
sub prettyFeature {

    my ($fStr,$wLookup) = @_;
    my $type;

    my ($feat,$nodeStr) = split(/$SEPARATOR/,$fStr);
    my @nodes = split(/ /,$nodeStr);

    if ($#nodes < 0) {
        return $fStr;
    }

    my $pNodeStr = $wLookup->{$nodes[0]};
    for (my $i=1; $i<=$#nodes; $i++){
        $pNodeStr .= " ".$wLookup->{$nodes[$i]};
    }

    if ($feat =~ /\((.*)\)=/) {
        my $match = $1;
        my @tlist = split(/ /,$match);
        my $targets = $wLookup->{$tlist[0]};
        for(my $i=1; $i<=$#tlist; $i++) {
            $targets .= " ".$wLookup->{$tlist[$i]};
        }
        $feat =~ s/\($match\)=/($targets)=/;
    } 

    return($feat.$SEPARATOR.$pNodeStr);
}

#
# Prints a feature match result in 2 columns: if a match is found, the feature
# prints in the middle.  If a feature is missed, it prints on the left, if an
# extra feature is provided, it prints on the right.
#
sub printMatch {

    my ($feature, $score, $matchType) = @_;

    my $fLen = length($feature);
    my $pStr;
    my $buf;
    my $scoreLen = 1;

    if ($score==0.5) {
        $scoreLen = 3;
    }

    if ($feature eq "ColumnLabels") {
        $pStr = "MatchType: Feature";
        $buf = " " x ($FEATURE_WIDTH - length($pStr));
        $pStr .= $buf;

        $pStr .= "  ";
        $buf = "G".(" " x int(($SCORE_WIDTH-2)/2))."M";
        $pStr .= $buf . (" " x (($SCORE_WIDTH - length($buf)) - 1)) . "S";
    } elsif ($matchType==$MISSED) {
        $pStr = "  Missed: ".$feature;
        $buf = " " x ($FEATURE_WIDTH - length($pStr));
        $pStr .= $buf . "  " . $score;
    } elsif ($matchType==$MATCHED) {
        $pStr = " Matched: ".$feature;
        $buf = " " x ($FEATURE_WIDTH - length($pStr));
        $pStr .= $buf . "  ";

        $buf = " " x int(($SCORE_WIDTH - $scoreLen) / 2);
        $pStr .= $buf . $score;
    } elsif ($matchType==$EXTRA) {
        $pStr = "   Extra: ". $feature;
        $buf = " " x ($FEATURE_WIDTH - length($pStr));
        $pStr .= $buf . "  ";

        $buf = " " x ($SCORE_WIDTH - $scoreLen);
        $pStr .= $buf . $score;
    }
    
    print $pStr."\n";
}

#
# auxiliary to printFeature: prints sums of each of the score columns;
#
sub printScoreTotal {

    my ($g, $m, $s) = @_;
    my $pstr = " " x $FEATURE_WIDTH;
    $pstr .= "  ";
    $pstr .= ("-" x $SCORE_WIDTH)."\n";
    $pstr .= (" " x ($FEATURE_WIDTH-6))."Total:  ";
    my $buf = "$g$m$s";
    $buf = $g.(" " x (($SCORE_WIDTH-length($buf))/2)).$m;
    $pstr .= $buf . (" " x ($SCORE_WIDTH-length("$buf$s"))) . $s;

    print $pstr."\n";
}

#
#  CreateFELookup( $fileName ) - reads frames.xml (as arg) and
#     returns a hash with core-FEs defined as keys
#
sub CreateFELookup {

    my $fileName = shift(@_);
    my %coreLookup;

    # initialize parser and read the file
    my $parser = new XML::Parser( Style => 'Tree' );
    my $tree = $parser->parsefile( $fileName );

    my @frames = @{$tree->[1]};
    
    # read content of <frames>
    for (my $f=0; $f<=$#frames; $f++) {

	if ($frames[$f] eq 'frame') {
	    $f++;
	    my @frame = @{$frames[$f]};
	    my %fatts = %{$frame[0]};
	    my $frName = $fatts{name};

	    print "parsing frame $frName ...\n" if $DEBUG;

	    # read content of <frame>
	    for (my $e=0; $e <= $#frame; $e++) {
		if ($frame[$e] eq 'fes') {
		    $e++;
		    my @fes = @{$frame[$e]};

		    # read content of <fes>
		    for (my $s=0; $s<=$#fes; $s++) {
			
			# each fe
			if ($fes[$s] eq 'fe') {
			    $s++;
			    my %featts = %{$fes[$s]->[0]};
			    my $feName=$featts{name};
			    my $feType=$featts{coreType};
			    print "FE $feName is of type $feType\n" if $DEBUG;
			    if ($feType eq 'Core') {
				# add lookup for Core FEs
				$coreLookup{$frName.'='.$feName} = 1;
			    }
			} # each fe
		    } # <fes>
		}
	    } # <frame>
	}
    } # <frames>

    return (%coreLookup);
}


#
# parseXMLFile ( filename ) - parses xml file and extracts information
#    relevant for scoring in a format conducive to comparison
#  
#   returns 3 references to hashes:
#      sentWords - concatenated content of <words> information
#                  a valid comparison can only be made if that
#                  information exactly matches.  (indexed by sent ID)
#      sentFeatures - indexed by sentence ID, each value is also a hash.
#                     each of those hashes is indexed by a feature composed of
#                     of Frame, FE, Supp, and NE names and their target/headwords,
#                     with the values being the score that each feature is worth.
#      sentWordLookup - indexed by sentence ID, each value is also a hash.
#                       each hash takes nodeIDs to their text values.
#
sub parseXMLFile {

    my $fileName = shift(@_);

    # initialize parser and read the file
    my $parser = new XML::Parser( Style => 'Tree' );
    my $tree = $parser->parsefile( $fileName );

    # list of sentences
    my @corpus = @{$tree->[1]};

    my %sentWordContent;
    my %sentFeatures;
    my %sentWordLookup;
    
    for ( my $i=0; $i <= $#corpus; $i++ ) {
        if ($corpus[$i] eq 'sentence') {
            $i++;
            my @sentence = @{$corpus[$i]};
            my %frFeatures;
            my %feFeatures;
	    my %suppFeatures;
            my %headWordLookup;
            my %wordLookup;

            print "found sentence ID=".$sentence[0]->{ID}."\n" if $DEBUG;
        
            for (my $j=1; $j <= $#sentence; $j++) {
                
                # Parse words--create wordLookup hash
		# words of compared sentences should be identical
		# create a string to compare
                if ($sentence[$j] eq 'words') {
                    $j++;
		    my $sentStr;
                    my @words = @{$sentence[$j]};
                    for (my $w=0; $w<=$#words; $w++) {
			# each word
                        if ($words[$w] eq "W") {
                            $w++;
                            my %watts= %{$words[$w]->[0]};
                            $wordLookup{$watts{ID}} = $watts{form};
                            $sentStr .= 'W'.$watts{ID}.$watts{start}
				.$watts{end}.$watts{form};
			    print "found word ".$watts{ID}."\n" if $DEBUG;
                            print "WORD $watts{ID} $wordLookup{$watts{ID}} \n" if $DEBUG;
                        }
                    }
		    $sentWordContent{$sentence[0]->{ID}} = $sentStr;
                }
                
                # Parse non-terminals
                if ($sentence[$j] eq 'non-terminals') {
                    $j++;
                    my @nts = @{$sentence[$j]};
                    for (my $n=0; $n <= $#nts; $n++) {

			# Each <P> Node
                        if ($nts[$n] eq 'P') {
                            $n++;
                            my @pnode = @{$nts[$n]};
                            my %patts = %{$pnode[0]};
                            my $frTarget;

                            print "found P node ".$patts{ID}."\n" if $DEBUG;
                            
			    # If <P> Node is a Frame:
			    #    create Frame Feature: Fname;Target worth FRAME_POINTS
			    #    create a headWordLookup entry for this node with Head=Target
			    #    if frame has a denoted FE, also create an FE Feature
                            if (defined $patts{Frame}) {
				#reformat Target: space separated
                                $frTarget = join(" ",sort(split(/, /,$patts{Target})));

                                # Target of a Frame is always a HEAD (also for NEs?)
                                $headWordLookup{$patts{ID}} = $frTarget;

				# Named Entities are counted differently from Frames
				if ($patts{Frame} =~ /^NE:/) {
				    $frFeatures{$patts{Frame}.$SEPARATOR.$frTarget} = $NE_POINTS;
				    print "NE $patts{Frame} $patts{Target}\n" if $DEBUG;
				} else {
				    $frFeatures{"Fr:".$patts{Frame}.$SEPARATOR.$frTarget} = $FRAME_POINTS;
				    print "Frame $patts{Frame} $patts{Target}\n" if $DEBUG;

				    # Denoted FEs (there can be multiple--comma separated)
				    if (defined $patts{Denoted}) {
					my @denotedFEs = split(/, /,$patts{Denoted});
					my $dfe;
					foreach $dfe (@denotedFEs) {
					    $feFeatures{"FE:".$patts{Frame}."($frTarget)".'='.$dfe.$SEPARATOR.$frTarget} = 1;
					}
				    }
				}
                            }
                            
                            # parse E Tags
                            my $headFound = 0;
                            my $semHeadFound = 0;
                            my $eRefs;
                            for (my $p=0; $p<$#pnode; $p++) {

                                # EACH E NODE:
                                if ($pnode[$p] eq 'E') {
                                    $p++;
                                    my @enode=@{$pnode[$p]};
                                    my %eatts = %{$enode[0]};

                                    # record list of all Refs to use as headlist in cases where no head is marked
                                    if (defined $eatts{Ref}) {
                                        if ($eRefs) {
                                            $eRefs .= " ".$eatts{Ref};
                                        } else {
                                            $eRefs = $eatts{Ref};
                                        }
                                    }
                                    
                                    # if frame element, then record in feFeatures
                                    if (defined $eatts{FE}) {
                                        $feFeatures{"FE:".$patts{Frame}."($frTarget)".'='.$eatts{FE}.$SEPARATOR.$eatts{Ref}} = 1;
                                        print "  FE $eatts{FE} $eatts{Ref}\n" if $DEBUG;
                                    }
                                    
                                    # Record in headWordLookup for the <P> containing this element
                                    if ($eatts{Head} eq "True" || $eatts{SemHead} eq "True") {

                                        if ($eatts{Head} eq "True") {
                                            $headFound = 1;
                                        } else {
                                            $semHeadFound = 1;
                                        }
                                        if (defined $headWordLookup{$patts{ID}}) {
                                            $headWordLookup{$patts{ID}} .= " ".$eatts{Ref};
                                        } else {
                                            $headWordLookup{$patts{ID}} = $eatts{Ref};
                                        }
                                    }

				    # If we are looking at an <E> inside a Supp <P>
				    # Record Supp features, all such <E>'s should be SemHeads

				    if (defined $patts{Supp}) {
					print STDERR "ERROR: non-SemHead <E> inside a Supp <P>\n" unless $eatts{SemHead} eq "True";
					if (defined $suppFeatures{"Supp".'='.$patts{Supp}}) {
					    $suppFeatures{"Su:Supp".$SEPARATOR.$patts{Supp}} .= " ".$eatts{Ref};
					} else {
					    $suppFeatures{"Su:Supp".$SEPARATOR.$patts{Supp}} = $eatts{Ref};
					}
				    }

                                } # end of each E node
                            }
                            # validity check for each P node
                            if (defined($patts{Frame}) && ($headFound + $semHeadFound) > 0) {
                                print STDERR "ERROR (sem)headwords found for non-transparent frame P ID=".$patts{ID}."\n"
                                    unless ($patts{Transparent} eq "True");
                            }
                            if (defined($patts{Frame}) && $patts{Transparent} eq "True" and $semHeadFound==0) {
                                print STDERR "ERROR no semheadwords found for transparent frame P ID=".$patts{ID}."\n";
                            }
                            if (!defined($patts{Frame}) && ($headFound + $semHeadFound)==0) {
                                print "No headwords are marked: using all edges as heads\n" if $DEBUG;
                                $headWordLookup{$patts{ID}} = $eRefs;
                            }
                            
                            print "HeadWords for ".$patts{ID}.": ".$headWordLookup{$patts{ID}}."\n\n" if $DEBUG;
                            
                        } # end of each P node
                    }
                    last; # bc. there's only 1 <non-terminals>
                }
            }

            # generate feature set

	    # frame features can be copied over
            my %matchSet = %frFeatures;

	    # for fe features, must resolve headwords using the lookups
            my $mStr;
            my $feFeat;
	    my $fe;
            my $feHWs;
            foreach $feFeat (keys %feFeatures) {
		($fe,$feHWs) = split(/$SEPARATOR/,$feFeat);
                $feHWs = join(" ",sort(split(/ /,&resolveHeadWords($feHWs,\%headWordLookup, \%wordLookup))));
                $mStr = $fe.$SEPARATOR.$feHWs;
		if (defined $CORELOOKUP{$fe}) {
		    $matchSet{$mStr} = $COREFE_POINTS;
		} else {
		    $matchSet{$mStr} = $NONCOREFE_POINTS;
		}
                print "Adding $mStr to matchSet\n" if $DEBUG;
            }

	    # for supp features
            my $supp;
	    my $suppHeads;
	    foreach $supp (keys %suppFeatures) {
		$suppHeads = join(" ",sort(split(/ /,&resolveHeadWords($suppFeatures{$supp},\%headWordLookup, \%wordLookup))));
		$mStr = $supp." ".$suppHeads;
		$matchSet{$mStr} = $SUPP_POINTS;
	    }

            #store in Sentence hash
            $sentFeatures{$sentence[0]->{ID}} = \%matchSet;
            $sentWordLookup{$sentence[0]->{ID}} = \%wordLookup;

        } # end of each sentence
    
    }
    return (\%sentWordContent, \%sentFeatures, \%sentWordLookup);
}

sub resolveHeadWords  {

    my ($str, $HwHash, $WHash) = @_;

    my @refs = split(/ /,$str);
    my @retList;
    my $ref;

    print "resolveHeadsCalled on $str\n" if $DEBUG;

    foreach $ref (@refs) {
        if ($WHash->{$ref}) {
            push(@retList, $ref);
        } elsif ($HwHash->{$ref}) {
            push(@retList, &resolveHeadWords($HwHash->{$ref}, $HwHash, $WHash));
        } else {
            print STDERR "ERROR: Cannot resolve HW for $ref--neither a P with a HW nor a W\n";
        }
    }

    my $retStr = join(" ",@retList);
    return $retStr;

}

sub usage {
    my $usage = <<USAGEEND;

usage: $PROG [options] <frames.xml> <goldstandard.xml> <file-to-score.xml>
     
   Script for scoring frame semantic structure extraction.
   Requires 3 arguments:

   <frames.xml>       : frXML/frames.xml file from FrameNet data.
                        This file can be downloaded from the
                        FrameNet public website at the following URL:
                        http://framenet.icsi.berkeley.edu/fnUsers/dataRequest

   <goldstandard.xml> : semXML containing extracted structure from gold
                        standard FrameNet full-text output XML.

   <file-to-score.xml>: file to score, semXML

   The script calculates (aggregated) recall and precision, and an
   fscore value calculated over all sentences in the file to score.

 Options:

   -h : displays this usage statement

   -v : produce verbose output, e.g. sentence text and information about which
        features matched and which did not.  Scores for missed features
        appear under "G", for matching features under "M", and for extra
        features (in the score file but not in the gold standard) under "S".
        The "totals" calculated for "G" and "S" include the sum of the
        points for matches, i.e. "M".

        Note: FE features also include the frame name, and the two are
        concatenated with an '=' character, i.e. Frame=FE.

   -s : print recall, precision, and fscore for each sentence 
        (by default, these are only printed for the file as a whole)

USAGEEND
		
    die $usage;
}
