#!/usr/local/bin/perl -w
#
#  File: fttosem.pl
#  Author: Adam Janin <adam@janin.org>
#          Michael Ellsworth <infinity@icsi.berkeley.edu>
# 
#  Copyright 2006, 2007 International Computer Science Institute
#
# Given a full-text XML file, convert first to DG, then output
# according to the Frame Semantic Structure Extraction requirements.
#
# 12/13/06  Initial Version
#           Adam Janin
#
#

use strict;
use FileHandle;
use Getopt::Std;
use SynSemTreeMaker;

use vars qw($opt_d $opt_v $opt_h $opt_o);

my($Debug, $Verbosity, $FileName);
my $XMLHeader="<?xml version=\"1.0\" encoding=\"UTF-8\"?>";

$opt_v = 0;
getopts("dv:ho:") || usage();

if ($#ARGV != 0 || defined($opt_h)) {
    usage();
}

$FileName = shift(@ARGV);

if (defined($opt_d)) {
    $Debug = 1;
} else {
    $Debug = 0;
}

if (!defined($opt_o)) {
    $opt_o=$FileName;
    $opt_o =~ s/xml$/sem/;
}

$Verbosity = $opt_v;


my $semfh = new FileHandle(">$opt_o") or die "Couldn't write $opt_o: $!";
    
print $semfh "$XMLHeader\n\n";
print $semfh "<corpus src=\"$FileName\">\n";

ProcessDGs($FileName, \&printsem, $Debug);

print $semfh "</corpus>\n";

$semfh->close();

print STDERR "Finished.\n";

# All subs below here

exit;



sub printsem {
    my($topnode, $errorcode) = @_;
    my($fname, $node, $edge, $childnode);
    my($gotmarker, $gothead);

# This commented-out code prevents the output of obviously bad trees.
# For the purposes of greater robustness, we're omitting it.

#    if ($errorcode != 0) {
#	FreeDG($topnode);
#	return;
#    }

    print $semfh "  <sentence ID=\"$topnode->{ID}\">\n";

    # Output word nodes
    print $semfh "    <words>\n";
    foreach $node (FindNodes($topnode, sub { $_[0]->{type} eq 'Word';})) {
	if ($node->{form} eq '&') {
	    print $semfh "      <W ID=\"$node->{sym}\" start=\"$node->{start}\" end=\"$node->{end}\" form=\"\&amp;\" />\n";
	} else {
	    print $semfh "      <W ID=\"$node->{sym}\" start=\"$node->{start}\" end=\"$node->{end}\" form=\"$node->{form}\" />\n";
	}
    }

    print $semfh "    </words>\n";

    print $semfh "\n    <non-terminals>\n";

    # Output frame nodes

    foreach $node (FindNodes($topnode, sub {$_[0]->{type} eq 'Frame';})) {
	print $semfh "\n      <P ID=\"$node->{sym}\" Frame=\"$node->{frameName}\"";

	my @targets=map {$_->{sym}} sort {$a->{start} <=> $b->{start}} @{$node->{targetNodes}};
	print $semfh " Target=\"", join(", ", @targets), "\"";

	if (grep {$_->{type} eq 'SemHead'} @{$node->{childEdges}}) {
	    print $semfh " Transparent=\"True\"";
	}

	if (exists($node->{denoted}) && ref($node->{denoted}) eq 'ARRAY' && $#{$node->{denoted}}>=0) {
	    print $semfh " Denoted=\"", join(", ", @{$node->{denoted}}), "\"";
	}
	print $semfh " >\n";

	foreach $edge (@{$node->{childEdges}}, @{$node->{nlChildEdges}}) {
	    $childnode = $edge->{childNode};
	    if ($edge->{type} eq 'Target') {
		next;
#		print $semfh "       <E Target=\"True\" Ref=\"$childnode->{sym}\" />";
#		printcomment($semfh, $childnode);
	    } elsif (exists($edge->{FE})) {
		if ($edge->{type} eq 'SemHead') {
		    print $semfh "       <E SemHead=\"True\" FE=\"$edge->{FE}\" Ref=\"$childnode->{sym}\" />";
		} else {
		    print $semfh "       <E FE=\"$edge->{FE}\" Ref=\"$childnode->{sym}\" />";
		}
		printcomment($semfh, $childnode);
		
#	    } elsif (exists($edge->{type}) && $edge->{type} eq 'NonHead') {
#		next;
#
	    } else {
		print $semfh "       <E Ref=\"$childnode->{sym}\" />";
		printcomment($semfh, $childnode);
	    }
	}
	foreach $edge (@{$node->{nlParentEdges}}, ($node->{parentEdge})) {
	    my $parentnode = $edge->{parentNode};
	    if (exists($edge->{reversedEdge}) && exists($edge->{reversedEdge}{FE})) {
		print $semfh "       <E FE=\"$edge->{reversedEdge}{FE}\" Ref=\"$parentnode->{sym}\" />";
		printcomment($semfh, $parentnode);
	    }		
	}
	print $semfh "      </P>\n";
    }

    # List
    foreach $node (FindNodes($topnode, sub { $_[0]->{type} eq 'List';})) {
	print $semfh "\n      <P ID=\"$node->{sym}\" >";
	printcomment($semfh, $node);
	foreach $edge (@{$node->{childEdges}}, @{$node->{nlChildEdges}}) {
	    $childnode = $edge->{childNode};
	    if ($edge->{type} eq 'Head') {
		print $semfh "       <E Head=\"True\" Ref=\"$childnode->{sym}\" />";
		printcomment($semfh, $childnode);
	    } elsif ($edge->{type} eq 'SemHead') {
		print $semfh "       <E SemHead=\"True\" Ref=\"$childnode->{sym}\" />";
		printcomment($semfh, $childnode);
	    } elsif ($edge->{type} eq 'NonHead' || $edge->{type} eq 'Relativized') {
#		next;
		print $semfh "       <E Ref=\"$childnode->{sym}\" />";
		printcomment($semfh, $childnode);
	    } else {
		print $semfh "       <E Ref=\"$childnode->{sym}\" />";
		printcomment($semfh, $childnode);
#		print STDERR "Node $node->{sym} edge $edge->{sym} unexpected type.\n";
#		next;
	    }
	}
	print $semfh "      </P>\n";
    }

    # MarkerPhrase
    # Marker edge gets no label; the non marker gets head. 
    # MarkerPhrase should only have two children, the marker and the head.

    foreach $node (FindNodes($topnode, sub { $_[0]->{type} eq 'MarkerPhrase';})) {
	$gothead = 0;
	print $semfh "\n      <P ID=\"$node->{sym}\"";

	my @markers=map {$_->{childNode}{sym}} grep {$_->{type} eq 'Marker'} @{$node->{childEdges}};
	print $semfh " Marker=\"", join(", ", @markers), "\" >";
	printcomment($semfh, $node);
	foreach $edge (@{$node->{childEdges}}, @{$node->{nlChildEdges}}) {
	    $childnode = $edge->{childNode};
	    if ($edge->{type} eq 'Marker') {
		next;
#		if ($gotmarker) {
#		    print STDERR "Got two markers for $node->{sym}\n";
#		}
#		$gotmarker = 1;
#		print $semfh "       <E Marker=\"True\" Ref=\"$childnode->{sym}\" />";
#		printcomment($semfh, $childnode);
	    } elsif (
		exists($edge->{type}) && $edge->{type} eq 'NonHead' ) {
#		next;
		print $semfh "       <E Ref=\"$childnode->{sym}\" />";
		printcomment($semfh, $childnode);

	    } elsif ($edge->{type} eq 'SemHead') {
		if ($gothead) {
		    print STDERR "Got two heads in marker phrase $node->{sym}\n";
		}
		$gothead = 1;
		print $semfh "       <E SemHead=\"True\" Ref=\"$childnode->{sym}\" />";
		printcomment($semfh, $childnode);
	    }
	}
	if (!$gothead) {
	    print STDERR "Missing marker and/or head on node $node->{sym}\n";
	}
	print $semfh "      </P>\n";
    }


    # Supp/Cop node (SemHead with Head, Cop/Supp marked as such, all others unmarked)

    foreach $node (FindNodes($topnode, sub {  $_[0]->{type} eq 'Supp' ||
						  $_[0]->{type} eq 'Cop' 
					      })) {
	print $semfh "\n      <P ID=\"$node->{sym}\"";
	my @governors = map {$_->{childNode}{sym}} grep {$_->{type} eq $node->{type}} @{$node->{childEdges}};
	print $semfh " Supp=\"", join(", ", @governors), "\" >";
	printcomment($semfh, $node);
	foreach $edge (@{$node->{childEdges}}, @{$node->{nlChildEdges}}) {
	    $childnode = $edge->{childNode};
	    if ($edge->{type} eq 'SemHead') {
		print $semfh "       <E SemHead=\"True\" Ref=\"$childnode->{sym}\" />";
		printcomment($semfh, $childnode);
	    } elsif ($edge->{type} eq 'Cop' ||
		     $edge->{type} eq 'Supp') {
		next;
#		print $semfh "       <E $edge->{type}=\"True\" Ref=\"$childnode->{sym}\" />";
#		printcomment($semfh, $childnode);

	    } elsif (exists($edge->{type}) && $edge->{type} eq 'NonHead') {
		next;

	    } else {
		next;
#		print $semfh "       <E Ref=\"$childnode->{sym}\" />";
#		printcomment($semfh, $childnode);
	    }
	}
	print $semfh "      </P>\n";
    }
    
    # Gov Node (Gov, Ant, Ctrlr marked head)
    
    foreach $node (FindNodes($topnode, sub { $_[0]->{type} eq 'Gov' || 
						 $_[0]->{type} eq 'Ctrlr' ||
						 $_[0]->{type} eq 'Ant' })) {
	print $semfh "\n      <P ID=\"$node->{sym}\" >";
	printcomment($semfh, $node);
	foreach $edge (@{$node->{childEdges}}, @{$node->{nlChildEdges}}) {
	    $childnode = $edge->{childNode};
	    if ($edge->{type} eq 'SemHead') {
		print $semfh "       <E SemHead=\"True\" Ref=\"$childnode->{sym}\" />";
		printcomment($semfh, $childnode);
	    } elsif ($edge->{type} eq 'Ctrlr' ||
		     $edge->{type} eq 'Gov' ||
		     $edge->{type} eq 'Ant') {
		print $semfh "       <E Head=\"True\" Ref=\"$childnode->{sym}\" />";
		printcomment($semfh, $childnode);

#	    } elsif (exists($edge->{type}) && $edge->{type} eq 'NonHead') {
#		next;

	    } else {
		print $semfh "       <E Ref=\"$childnode->{sym}\" />";
		printcomment($semfh, $childnode);
	    }
	}
	print $semfh "      </P>\n";
    }

    print $semfh "    </non-terminals>\n";
    print $semfh "  </sentence>\n";

    FreeDG($topnode);
    print STDERR "\nWrote sentence $topnode->{ID} \n\n";

} # printsem()

sub printcomment {
    my($semfh, $node) = @_;
    if ($node->{type} eq 'Word') {
	if ($node->{form} eq '--') {
	    print $semfh "  <!-- &#150; -->";
	} elsif ($node->{form} eq '&') {
	    print $semfh "  <!-- \&amp; -->";
	} else {
	    print $semfh "  <!-- $node->{form} -->";
	} 
    } elsif (exists($node->{PT})) {
	print $semfh "  <!-- $node->{PT} -->";
    }

    print $semfh "\n";
}

sub usage {
    print STDERR <<EndOfUsage;

Usage: fttosem.pl [-d] [-v \#] [-o outfile] file.xml
  (Currently, -v is not supported)

  Convert a full-text XML file to a form appropriate for the Frame
  Semantic Structure Extraction task. The output is one file with <sentence>
  </sentence> around the each sentence in the input. If the input file is 
  file.xml, the output file is file.sem unless otherwise specified with -o.

EndOfUsage
    exit;
} # usage()
